/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of International Business Machines may not be used to endorse 
 *     or promote products derived from this software without specific prior 
 *     written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR 
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR 
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR 
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Alternatively, provided that this notice is retained in full, this
 * software may be distributed under the terms of the GNU General
 * Public License ("GPL") as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version,
 * in which case the provisions of the GPL apply INSTEAD OF those given above.
 *
 ****************************************************************************/
 static char sccsid[] = "@(#)37	1.115.1.47  src/avs/fs/mmfs/ts/kernext/gpl-linux/lxtrace.c, mmfs, avs_rhrz, rhrzs024a 2/17/15 18:58:24";
/**************************************************************************
 *
 * lxtrace : the command.  Turns tracing on/off and formats trace files.
 *
 *  "lxtrace {on hw[,hw[,hw...]] [trcFile [filesize [bufsize]]]" }
 *           {off                                               }
 *           {format [-T] [-v] [-t formatFile] [-o outFile] [trcFile] }
 *
 *           on   - begins tracing
 *           off  - stops tracing
 *           hw   - hookword (identifies traces to be captured)
 *                  This is a comma-separated list (up to LXTRACE_MAX_HW)
 *           trcFile  - path to the file where output is written
 *           filesize - maximum size (bytes) of the wrap-around file
 *                      A filesize of 0 will cause the file NOT to wrap
 *           bufsize  - the size of buffer to used by the device
 *                      to record trace records
 *
 *   For debug/testing purposes, the following additional options
 *   are defined:
 *   "lxtrace dump              // returns state information in a buffer
 *           fsync              // flush buffered records to the read buffer
 *
 *
 **************************************************************************/
#ifdef GPFS_LINUX
/* Added for LFS */
#define _LARGEFILE_SOURCE
#define _FILE_OFFSET_BITS 64
#include <Shark-gpl.h>
#else
#include <cxiTypes.h>          // needed by windows per Demyn
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/mman.h>

#include <sys/ipc.h>
#include <sys/msg.h>
#include <sys/stat.h>
#include <dlfcn.h>
#include <limits.h>

#include <Trace.h>
#define ROUNDUP(_addr, _align) \
   (((long)_addr + (long)_align - 1L) & ~((long)_align - 1L))
#include <lxtrace.h>
#ifndef O_DIRECT
#ifdef GPFS_ARCH_PPC64
#define O_DIRECT 0400000
#else
#define O_DIRECT 040000
#endif
#endif

#if defined(GPFS_LINUX) && defined(LXTRACE_USE_RELAY)
extern int RelayDump(void);
extern int RelayStart(long long outSize, int bufSize, int bufNum,
                      int interval, int level);
#endif
static int (*uncompress)(char *dst, unsigned long *dlen, 
                         char *src, unsigned long slen);
extern int msgGet(int flags);
extern unsigned int cpuNum;
extern size_t compLevel;
extern int useDIO;
extern long long percpuOutSize;
extern size_t subBufSize;
extern int writeMode;
extern int noWrap;
extern char percpuOutBasename[PATH_MAX];
/* Examples of formatted trace output:

   Default options:

    Timestamp   Pid  COMPONENT_TAG: application trace record
   ----------- ----- ---------------------------------------
      0.000000 21499 TRACE_VNODE: gpfs_i_revalidate enter:

   -v option:

    Timestamp   Pid  P Hookword COMPONENT_TAG: application trace record
   ----------- ----- - -------- ---------------------------------------
      0.000000 21499 1 307004B3 TRACE_VNODE: gpfs_i_revalidate enter:

   -T option:

      Timestamp      Pid  COMPONENT_TAG: application trace record
   ---------------- ----- ---------------------------------------
   970706704.711164 21499 TRACE_VNODE: gpfs_i_revalidate enter:

   -T and -v options:

      Timestamp      Pid  P Hookword COMPONENT_TAG: application trace record
   ---------------- ----- - -------- ---------------------------------------
   970706704.711164 21499 1 307004B3 TRACE_VNODE: gpfs_i_revalidate enter:

 */

/* Define fragments of header lines and the corresponding trace printf format
   strings for each field on a trace output line */
#define H1_TS       " Timestamp  "
#define H2_TS       "----------- "
#define FMT_TS      "%4d.%06d "
#define H1_TS_T     "    Timestamp     "
#define H2_TS_T     "----------------- "
#define FMT_TS_T    "%10d.%06d "

#ifdef GPFS_SOLARIS
#define H1_PID      "  Pid  :  Tid  "
#define H2_PID      "-------------- "
#define FMT_PID     "%6d : %5lu "
#else
#define H1_PID      "  Pid  "
#define H2_PID      "------ "
#define FMT_PID     "%6d "
#endif

#define H1_SEQ      " Sequence  "
#define H2_SEQ      "---------- "
#define FMT_SEQ     "%10llu "

#ifdef SHARED_TRC_BUFF
#define H1_CYCLE    "CPU Cycle Count    Wrap "
#define H2_CYCLE    "----------------  ----- "
#define FMT_CYCLE   "%17lld %5d "
#endif

#define H1_TS_D     "           Date            "
#define H2_TS_D     "-------------------------- "
#define FMT_TS_D    "%19s.%06d "

#define H1_P_HW     ""
#define H2_P_HW     ""
#define FMT_P_HW    ""
#define H1_P_HW_V   "P B Hookword "
#define H2_P_HW_V   "- - -------- "
#define FMT_P_HW_V  "%1d %1d %08X "

#define H1_TAG      "COMPONENT_TAG: application trace record\n"
#define H2_TAG      "---------------------------------------\n"
#define FMT_TAG     "%s: "

/* Bytes in an "_id2name" formatting substitution */

#define FMT_TRC_BEGIN "Trace begins: %s\n"
#define FMT_TRC_END "Trace ends  : %s%s\n"
#define FMT_COMP_RATIO "Trace compression ratio: %2d\n"
#define FMT_LOST_COUNT "Trace records lost in total: %10llu\n"

#define TRC_BEGIN_STR "formatting...           "
#define TRC_END_STR   "formatting...           ","               "
#define FILE_WRAP_STR " (file wrapped)"
#define FILE_NOWRAP_STR "               "
#ifndef MIN
#define MIN(a,b)  ((a)<(b) ? (a) : (b))
#endif

#if defined( __64BIT__) || defined(_LP64)
typedef long long argType;
#define ARGLEN sizeof(long long)
#define TRC_ARG(N) (((long long *)&application_data)[N])
#else
typedef int argType;
#define ARGLEN sizeof(int)
#define TRC_ARG(N) (((int *)&application_data)[N])
#endif // __64BIT_ || _LP64_

#define TRC_ARGLIST argP[0] , argP[1] , argP[2] , argP[3] , argP[4] , argP[5], \
                    argP[6] , argP[7] , argP[8] , argP[9] , argP[10], argP[11],\
                    argP[12], argP[13], argP[14], argP[15], argP[16], argP[17],\
                    argP[18], argP[19]
#define TRC_ARGLIST1F argP[0], *d
#define TRC_ARGLIST2F argP[0], argP[1], *d
#define TRC_ARGLIST3F argP[0], argP[1], argP[2], *d
#define TRC_ARGLIST4F argP[0], argP[1], argP[2], argP[3], *d
#define TRC_ARGLIST5F argP[0], argP[1], argP[2], argP[3], argP[4], *d
#define TRC_ARGLIST6F argP[0], argP[1], argP[2], argP[3], argP[4], argP[5], *d
#define TRC_ARGLIST7F argP[0], argP[1], argP[2], argP[3], argP[4], argP[5], \
                      argP[6], *d
#define TRC_ARGLIST8F argP[0], argP[1], argP[2], argP[3], argP[4], argP[5], \
                      argP[6], argP[7], *d
#define TRC_ARGLIST9F argP[0], argP[1], argP[2], argP[3], argP[4], argP[5], \
                       argP[6], argP[7], argP[8], *d
#define TRC_ARGLIST10F argP[0], argP[1], argP[2], argP[3], argP[4], argP[5], \
                       argP[6], argP[7], argP[8], argP[9], *d


/* Offsets of the various fields in the formatting file */
#define GFR_TRCID 0
#define GFR_HOOKID 15
#define GFR_COMPID 24
#define GFR_FORMAT 44

/* Max length of COMPID in the formatting file */
# define MAX_COMPID_LEN (GFR_FORMAT - GFR_COMPID - 1)

/* Thread IDs greater than this are really message handlers.
   FirstMsgHandlerThread is also defined in ThreadNames.h. */
#define FirstMsgHandlerThread 0x200

/* Size of the hash tables for in-memory format file and its hash functions.
   HOOK_HASH is for looking up format strings by hook id.
   I2N_HASH is for looking up id2name conversions by conversion name and id
   value; the I2NHASH hash function uses the first 4 chars of the conversion
   name, relying on the fact that these names are at least 4 chars long,
   because they all end in "_id2name". */
#define HOOK_HASH_TABLE_SIZE 0x100000
/* The full hookid values for hookhash are of the form YYY0XXXX, where Y is the
   sharkhookid and X is the trace id. The sharkhookids are one of 30[6-9], so only
   the last digit of the hookid is important. */
int hookhash(int H)
{
  int hookid = ((H & 0x00F00000) >> 4); //we don't need the 0 in the middle
  int trcid = (H & 0x0000FFFF);
  return (hookid | trcid);
}
#define I2N_HASH_TABLE_SIZE 0x4000
#ifndef ALIGNED_DATA_ACCESS
#define I2NHASH(F,I) ((*(int*)(F) + (I)) & 0x3FFF)
#else
static inline int I2NHASH(char *F, int I)
{
  int tmp;

  memcpy(&tmp, F, sizeof(int));
  return (tmp + I) & 0x3FFF;
}
#endif

/* combined device and application trace data headers */
typedef struct trc_hdrs_t
{
  trc_header_t hdr;  /* device header */
  trc_datahdr_t rec; /* application header */
} trc_hdrs_t;

/* Each format record has one of these headers describing it. */
typedef struct fmtDesc_t
{
  int       hookid;             // hookid or id2name value
  char      *cidP;              // component identifier or id2name tag
  char      *fmtP;              // format string or id2name string
  i2nSpec_t *i2nSpecP;          // array of id2name conversion specifiers
  struct fmtDesc_t *fmtNextP;   // next in the hash chain
  int numHits;                  //number of times we see this hookid
  Int64 totalSpace;             //total space in the trcfile for this hookid
  int largestEntry;             //the largest entry (by space) for this hookid
                                //in the trcfile
} fmtDesc_t;


/* Hash tables for finding format records. */
fmtDesc_t * hookHashP[HOOK_HASH_TABLE_SIZE];
fmtDesc_t * i2nHashP[I2N_HASH_TABLE_SIZE];

typedef struct I2NValueList
{
  int nValues;                  // number of hookid values found
  int nAllocated;               // size of the values array
  char *cidP;                   // id2name tag
  UInt64 *valuesP;              // array of hookid values
  struct I2NValueList *vlNextP; // next ptr in the hash table
} I2NValueList;

/* Hash table for finding all hookid values for a given id2name tag/cid. */
I2NValueList * i2nValueListHashP[I2N_HASH_TABLE_SIZE];

/* Allocate a new structure and initialize its members */
I2NValueList *newI2NValueList(char *newCidP)
{
  I2NValueList *newVLP = NULL;
  if ((newVLP = (I2NValueList *)malloc(sizeof(I2NValueList))) == NULL)
    return NULL;

  newVLP->cidP = newCidP;
  newVLP->nValues = 0;
  newVLP->nAllocated = 0;
  newVLP->valuesP = NULL;
  newVLP->vlNextP = NULL;
  return newVLP;
}

/* Add a new value to the values array, create the
   array or increase its size if necessary */
int addValue(I2NValueList* vlP, UInt64 value)
{
  /* Increase the size of the values array if needed */
  if (vlP->nValues == vlP->nAllocated)
  {
    UInt64 *oldvaluesP = vlP->valuesP;
    vlP->nAllocated += 64;

    if (NULL ==
        (vlP->valuesP = (UInt64 *)malloc(vlP->nAllocated * sizeof(UInt64))))
      return ENOMEM;

    if (vlP->nValues > 0 && oldvaluesP != NULL)
    {
      memcpy(vlP->valuesP, oldvaluesP, vlP->nValues*sizeof(UInt64));
      free(oldvaluesP);
    }
  }

  /* add the new value to the values array */
  vlP->valuesP[vlP->nValues++] = value;
  return 0;
}

/* find an entry in the I2NValueList hash table, create if necessary */
int findI2NValueListInHashTable(char *cidP, I2NValueList **curI2NVLPP,
                                int createIfNotFound)
{
  I2NValueList *nextP = NULL;
  int hashVal = I2NHASH(cidP, 0); // for the id2name tag/cid

  /* search hash chain for the specified tag/cid */
  for (nextP = i2nValueListHashP[hashVal];
       nextP != NULL;
       nextP = nextP->vlNextP)
  {
    if (strcmp(nextP->cidP, cidP) == 0)
    {
      *curI2NVLPP = nextP; // an existing value list is found
      return 0;
    }
  }

  /* A value list with the same cid is not found */
  if (createIfNotFound == 0)
  {
    *curI2NVLPP = NULL;
    return ENOENT;
  }

  /* create the new value list */
  if ((nextP = newI2NValueList(cidP)) == NULL)
    return ENOMEM;

  /* add the new value list to the hash table */
  nextP->vlNextP = i2nValueListHashP[hashVal];
  i2nValueListHashP[hashVal] = nextP;
  *curI2NVLPP = nextP;

  return 0;
}

/* Returns a positive number if the 2nd value is larger to allow
   the values to be sorted in descending numerical order */
int compareHookid(const void *p1, const void *p2)
{
  if ( *(UInt64*)p2 > *(UInt64*)p1 )
    return 1;
  else if ( *(UInt64*)p2 < *(UInt64*)p1 )
    return -1;
  else
    return 0;
}

void sortValues(I2NValueList *vlP)
{
  qsort (vlP->valuesP, vlP->nValues, sizeof(UInt64), compareHookid);
}

void sortI2NValueLists()
{
  int l, j;
  I2NValueList *next = NULL;
  
  for (l = 0; l < I2N_HASH_TABLE_SIZE; l++)
    for (next = i2nValueListHashP[l]; next != NULL; next = next->vlNextP)
      sortValues(next);

#ifdef DEBUG_I2N_VL
  for (l = 0; l < I2N_HASH_TABLE_SIZE; l++)
    for (next = i2nValueListHashP[l]; next != NULL; next = next->vlNextP)
    {
      fprintf(stderr, "I2NValueLists: l=%d cidP=%s nValues=%d nAllocated=%d\n",
              l, next->cidP, next->nValues, next->nAllocated);
      for (j = 0; j < next->nValues; j++)
        fprintf(stderr, "0x%lX ", next->valuesP[j]);
      fprintf(stderr, "\n");
    }
#endif
}

/* Hash lookup statistics */
static int nHookLookup = 0;
static int nHookCheck = 0;
static int nI2NLookup = 0;
static int nI2NCheck = 0;
static int nI2NFail = 0;

/* static state information */
int TraceHandle = -1; // file descriptor for the trace device
static long long outSize;       // out_fd filesize
static Int64 bufSize;       // size of the trace buffers (device and daemon)
static int bufNum;
static int interval;
static int useRelay;
static char *bufP;        // the daemon trace read buffer
static int hdrSize[LXTRACE_NR_CPUS];  // trace header size, for backward compatibility
unsigned long long totalPostCompSize = 0;
unsigned long long totalTrcfileProcessedSize = 0; //I need to track this seperately since overwrite mode doesn't do compression
unsigned long long totalCompSize = 0;

#if !defined(GPFS_SOLARIS)
/* this code is used on Linux and Solaris, but only allow DIO on Linux */
#define LXTRACE_USE_DIO
#endif
static int nWraps;        // number of times the output file has wrapped
static off_t currentPos;  // position of next write to output file
static off_t fileEnd=0;  // position of next write to output file
static int curCPU;
extern int msgQueue;
extern int relayBuf;
static int genStats = 0;
/* The previous time is used to validate a wrap point when we
   think we've encountered one. */
static struct timeval prevTime[LXTRACE_NR_CPUS];
struct timeval tzero = { 0, 0 }; /* Used for the first trace entry */
/* Keep track of start/end times in the trace file */
static time_t startTime, stopTime;
extern Boolean lxtraceDebug;

#ifndef GPFS_WINDOWS
static char *formatFileP;     // malloc'ed buffer holding formatting data
static char *fmtDescSpaceP;   // fmtData_t space for each format record
#else
extern char *formatFileP;     // malloc'ed buffer holding formatting data
extern char *fmtDescSpaceP;   // fmtData_t space for each format record
#endif

static int allowed_usec_wobble = 2; // allowed timestamp wobble before declaring a wrap.

#ifdef SHARED_TRC_BUFF
char *TODFmtP = "The current timestamp is %lld.%lld.\n";
char *TODcidP = "TRACE_TIME";
char *dynamicTODMsg = "Using TODs to calculate cpu speed for timestamps\n";
char *staticTODMsg = "Using invariant time cycle counter at %fMhz to calculate timestamps\n";
Int64 wrapOffset;
int wrapNumber;
double tscSpeed = -1; /* If tscSpeed is -1 (ie not xlinux) then we use approximations */

typedef struct refTOD
{
  long long cycleCount;
  struct timeval timestamp;
  double cpuSpeed;
} refTOD;

static struct timeval zeroTime = { 0, 0};
int currTOD;
int firstTOD = 1; /* Traces before the first TOD still use the first TOD */
struct refTOD **TODList;
int TODMaxListSize = 4096;
int currTODCount = 0;
int showCycleCountOnly = 0; /* Only used if we don't have the cpu frequency and TODs */
int showCycleCount = 0;

struct timeval subtrTimeval (struct timeval old, struct timeval new)
{
  /* Note it's possible for "new" to be a negative time */
  struct timeval result;
  if ( old.tv_usec >= new.tv_usec )
  {
    result.tv_usec = old.tv_usec - new.tv_usec;

    if (result.tv_usec > 1000000)
    {
      old.tv_sec++;
      result.tv_usec -= 1000000;
    }
  }
  else
  {
    old.tv_sec -= 1;
    result.tv_usec = (old.tv_usec + 1000000) - new.tv_usec;
  }
  result.tv_sec = old.tv_sec - new.tv_sec;
  return result;
}

/* We pick a particular TOD timestamp to do the calculations from */
struct timeval calculateTimestamp(long long refCycleCount, struct timeval refTime,
                                  long long newCycleCount, double cpuFreq)
{
  /* I'm not sure how to handle the case where the cycle count wraps, so lets assume it won't
           for the moment. */
  long long cycleDiff = refCycleCount - newCycleCount;
  long long timeDiff = cycleDiff / cpuFreq; /* timeDiff is in microseconds */
  struct timeval diffTime, newTime;
  newTime.tv_sec = timeDiff / 1000000;
  newTime.tv_usec = timeDiff % 1000000;

  diffTime = subtrTimeval(refTime, newTime);
      DFP("refCycle %lld newCycle %lld cycleDiff %lld timeDiff %lld (%lld.%lld) cpuFreq %f refTime %lld.%lld resultTime %10d.%06d\n",
         refCycleCount, newCycleCount, cycleDiff, timeDiff, newTime.tv_sec, newTime.tv_usec, cpuFreq, refTime.tv_sec, refTime.tv_usec, diffTime.tv_sec, diffTime.tv_usec);
  return diffTime;
}
#endif

/* Convert ascii representation of a hexadecimal to int */
int axtoi(char s[])
{
  int i, n;

  n = 0;
  for (i = 0; s[i] != '\0'; i++)
  {
    if (s[i] >= '0' && s[i] <= '9')
      n = 16 * n + (s[i] - '0');
    else if (s[i] >= 'A' && s[i] <= 'F')
      n = 16 * n + (s[i] - 'A' + 10);
    else if (s[i] >= 'a' && s[i] <= 'f')
      n = 16 * n + (s[i] - 'a' + 10);
    else
      break;
  }
  return n;
}


/* Parse a quoted format string: removes enclosing quotes, removes trailing
   white-space, and converts escape sequences \n, \r, \t, \", etc.
   It's ok to call this function to convert a string in-place (outP == inP),
   since the conversion only makes the string shorter (removes surrounding
   quotes and reduces escape sequences to a single char).
   Returns a pointer to the next character in inP following the ending quotes,
   or NULL if there are no more characters following the quoted string. */
char *ParseFormatString(char *inP, char *outP)
{
  char *initialOutP = outP;
  int b = 1;

  /* skip initial quote */
  if (*inP == '"')
    inP++;

  /* copy string until ending quote */
  while (*inP && *inP != '"')
  {
    if (*inP == '\\' && *(inP + 1))
    {
      inP++;
      switch (*inP)
      {
        case 'n': *outP++ = '\n'; inP++; break;
        case 'r': *outP++ = '\r'; inP++; break;
        case 't': *outP++ = '\t'; inP++; break;
        default:  *outP++ = *inP++;
      }
    }
    else if (*inP == '%')
    {
      *outP++ = *inP++;
      if (*inP == '%')
        *outP++ = *inP++;
      else
      {
        if (*inP == '+' || *inP == '-')
          *outP++ = *inP++;
        while (isdigit(*inP))
          *outP++ = *inP++;
        b <<= 1;
      }
    }
    else
      *outP++ = *inP++;
  }

  /* remove trailing white space from the string */
  while (outP > initialOutP && isspace(outP[-1]))
    outP--;
  *outP = '\0';

  /* Return pointer to next character after the string or NULL if no ending
     quote was found */
  return *inP? inP + 1: NULL;
}
#ifdef SHARED_TRC_BUFF
/* Figure out if we have a trcfile from overwrite (new) or blocking (old) mode */
int checkTraceFileMode(char *trcPrefixP)
{
  char trcFileP[PATH_MAX];
  struct stat buf;
  sprintf(trcFileP, "%s.cpu%d", trcPrefixP, 0);
  int fd = open(trcFileP, O_RDONLY);
  int thisChar;
  int seqHeader = 0, sharedBuffHeader = 0;
  int bytesRead;
  int totalRead = 0;
  if (stat(trcFileP, &buf) == 0)
  {
    if (buf.st_size <= 0)
    {
      DP("%s is empty. No trace entries detected.\n", trcFileP);
      exit(1);
    }
  }
  while((bytesRead = read(fd, &thisChar, sizeof(int))) != 0)
  {
    if (thisChar == LXTRACE_MAGIC || thisChar == LXTRACE_SEQ_MAGIC || thisChar == LXTRACE_COMP_MAGIC)
      seqHeader++;
    if (thisChar == LXTRACE_SWAP_MAGIC ||  thisChar == LXTRACE_TOD_MAGIC)
      sharedBuffHeader++;
    /* There should never be a mix of headers, so stop after we find a few */
    if(abs(seqHeader - sharedBuffHeader) > 20)
      break;
  }
  if(sharedBuffHeader == seqHeader)
  {
    DP("Could not determine whether this used overwrite or blocking mode. No headers were detected\n");
    exit(1);
  }

  if (sharedBuffHeader > seqHeader)
    return LXTRACE_NB;
  return LXTRACE_BLK;
}
#endif
/* Parse a string containing a list of id2name conversion specifiers of the
   form "$n=xxx_id2name $m=yyy_id2name ...".  Returns a "null-terminted" array
   of i2nSpec_t structures, i.e., the array contains one more entry than the
   number of conversion specifiers found, and argNum/i2nNameP in the last
   array entry are set to 0/NULL. */
i2nSpec_t *ParseI2N(char *i2nP)
{
  i2nSpec_t i2n[LXTRACE_MAX_FORMAT_SUBS];
  i2nSpec_t *i2nSpecP;
  char *p;
  int n, argNum, a, nll;
  unsigned int m;

  for (n = 0, p = strtok(i2nP, " \t\n");
       p && n < LXTRACE_MAX_FORMAT_SUBS;
       n++, p = strtok(NULL, " \t\n"))
  {
    if (*p != '$')
      break;
    p++;
    argNum = atoi(p);
    p = strchr(p, '=');
    if (argNum <= 0 || p == NULL)
      break;

    i2n[n].argNum = argNum;
    i2n[n].i2nNameP = p + 1;
  }
  if (n == 0)
    return NULL;  // ?? print msg about bad id2name specifiers

  i2nSpecP = (i2nSpec_t *)malloc((n + 1)*sizeof(i2nSpec_t));
  if (i2nSpecP != NULL)
  {
    memcpy(i2nSpecP, i2n, n*sizeof(i2nSpec_t));
    i2nSpecP[n].argNum = 0;
    i2nSpecP[n].i2nNameP = NULL;
  }
  return i2nSpecP;
}

/* The formatting records are read into memory and parsed.  A fmtDesc_t header
   is constructed for each and is then inserted into a hash table. */
int BuildFormatTable(FILE *fP)
{
  char buf[2048];
  long fileSize;   /* size (bytes) of the formatting file */
  int  nRecords;   /* Number of records in the formatting file */
  int  hookid;     /* hookid of current record */
  int  hashVal;    /* hash value */
  struct fmtDesc_t *nextDescP;
  char *curFmtRecordP, *nextFmtRecordP;
  char *trcidP;
  char *hookP;
  char *cidP;
  char *fmtP;
#ifdef DBGASSERTS
  char *formatFileEndP;
  char *fmtDescSpaceEndP;
#endif // DBGASSERTS
  I2NValueList *curI2NVLP = NULL;
  int rc = 0;

  /* Clear all entries in the hash table */
  memset(hookHashP, 0, sizeof(hookHashP));
  memset(i2nHashP,  0, sizeof(i2nHashP));
  memset(i2nValueListHashP,  0, sizeof(i2nValueListHashP));

  rewind(fP);

  /* Count the number of formatting records */
  nRecords = 0;
  while (fgets(buf, sizeof(buf), fP) != NULL)
    nRecords += 1;

  /* Now that we are at the end of the file, see how big it is */
  fileSize = ftell(fP);
  if (fileSize < 0)
    return errno;

  /* Allocate the buffer for the file contents.
     Add 1 for possible trailing zero. */
  if (NULL == (formatFileP = (char *)malloc(fileSize+1)))
    return ENOMEM;

  /* Allocate space for the fmtData_t for all of the records. */
  if (NULL == (fmtDescSpaceP = (char *)malloc(nRecords * sizeof(fmtDesc_t))))
    return ENOMEM;
  nextDescP = (fmtDesc_t *)fmtDescSpaceP;

#ifdef DBGASSERTS
  formatFileEndP = formatFileP + fileSize;
  fmtDescSpaceEndP = fmtDescSpaceP + (nRecords * sizeof(fmtDesc_t));
#endif // DBGASSERTS

  rewind(fP); /* back to the beginning of the file */

  /* Read each record in one-at-a-time, parse the record, and insert it into
     the right hash table. */
  curFmtRecordP = formatFileP;
  while (fgets(curFmtRecordP, sizeof(buf), fP) != NULL)
  {
    nextFmtRecordP = curFmtRecordP + strlen(curFmtRecordP);

    /* The TRCID identifier, hookid, component id, and format string are at
       fixed offsets */
    trcidP = curFmtRecordP + GFR_TRCID;
    hookP  = curFmtRecordP + GFR_HOOKID;
    cidP   = curFmtRecordP + GFR_COMPID;
    fmtP   = curFmtRecordP + GFR_FORMAT;

    /* parse the format record. */
    hookid = axtoi(hookP) & 0xFFFFFFFF;
    nextDescP->hookid = hookid;
    nextDescP->cidP = strtok(cidP, " ");
    nextDescP->fmtP = fmtP;

    /* find the end of the format string */
    fmtP = ParseFormatString(fmtP, fmtP);

    /* check for id2name conversion specifiers following the format string */
    nextDescP->i2nSpecP = NULL;
    if (fmtP)
    {
      while (isspace(*fmtP))
        fmtP++;
      if (*fmtP)
        nextDescP->i2nSpecP = ParseI2N(fmtP);
    }

    /* Place the format record descriptor into the right hash table. */
    if (strncmp(trcidP, "TRCID_", 6) == 0)
    {
      /* this record contains a format string */
      strcat(nextDescP->fmtP, "\n");
      hashVal = hookhash(hookid);
      nextDescP->fmtNextP = hookHashP[hashVal];
      hookHashP[hashVal] = nextDescP;
      if (genStats)
      {
        hookHashP[hashVal]->numHits = 0;
        hookHashP[hashVal]->totalSpace = 0;
        hookHashP[hashVal]->largestEntry = 0;
      }
    }
    else
    {
      /* this record contains an id2name conversion value */
      hashVal = I2NHASH(nextDescP->cidP, hookid);
      nextDescP->fmtNextP = i2nHashP[hashVal];
      i2nHashP[hashVal] = nextDescP;

      /* If the current value list cannot be used, get a new one */
      if (curI2NVLP == NULL ||
          (strcmp(curI2NVLP->cidP, nextDescP->cidP) != 0))
      {
        if ((rc = findI2NValueListInHashTable(nextDescP->cidP,
                                              &curI2NVLP, 1)) != 0)
          return rc;
      }

      /* Add the new hookid value to the current value list */
      if ((rc = addValue(curI2NVLP, hookid)) != 0)
        return rc;
    }
    /* Advance to the next fmtDesc and format record */
    nextDescP++;
    curFmtRecordP = nextFmtRecordP;
  }

#ifdef DBGASSERTS
  /* verify that we didn't overrun our buffers */
  if ((char *)nextDescP > fmtDescSpaceEndP)
  {
    DP("%s: overwrote format descriptor memory!\n", TRCFMT_NAME);
    return -1;
  }
  if (nextFmtRecordP > formatFileEndP)
  {
    DP("%s: overwrote format file memory!\n", TRCFMT_NAME);
    return -1;
  }
#endif // DBGASSERTS

  sortI2NValueLists();

  return 0;
}

/* Find a format record by hookid. */
int GetFormatRecord(FILE *fP, int hookid, char **cidPP, char **fmtPP,
                    i2nSpec_t **i2nPP,
                    Int64 spaceUsed)
{
  fmtDesc_t *nextP;

  DFP("GetFormatRecord(0x%X, 0x%X, 0x%X, 0x%X, 0x%X)\n",
      fP, hookid, bufSize, bufP, cidPP);

  nHookLookup++;
  #ifdef SHARED_TRC_BUFF
  if(hookid == LXTRACE_TOD_MAGIC)
  {
    *cidPP = TODcidP;
    *fmtPP = TODFmtP;
    *i2nPP = NULL;
    DFP("GetFormatRecord for id %X is: %s\n", hookid, *fmtPP);
    return 0;
  }
#endif
  /* search hash chain for the specified hookid */
  for (nextP = hookHashP[hookhash(hookid)];
       nextP != NULL;
       nextP = nextP->fmtNextP)
  {
    nHookCheck++;

    /* Check whether hookid matches. */
    if (nextP->hookid == hookid)
    {
      /* Hookid and the record type match.  Copy to work area. */
      *cidPP = nextP->cidP;
      *fmtPP = nextP->fmtP;
      *i2nPP = nextP->i2nSpecP;
      if (genStats)
      {
        nextP->numHits++;
        nextP->totalSpace += spaceUsed;
        if (nextP->largestEntry < spaceUsed)
          nextP->largestEntry = spaceUsed;
        totalTrcfileProcessedSize += spaceUsed;
      }
      DFP("GetFormatRecord for id %X is: %s\n", hookid, *fmtPP);
      return 0;
    }
  }

  /* not found */
  return -1;
}

/* For genStats (--stat flag), sort the hookids for display */
int hookhashComparator(const void *p1, const void *p2)
{
  fmtDesc_t *a = *((fmtDesc_t  **) p1);
  fmtDesc_t *b = *((fmtDesc_t **) p2);

  if (a == NULL && b == NULL)
    return 0;
  else if (a == NULL)
    return -1;
  else if (b == NULL)
    return 1;

  //sort first by hits, then by total space
  if (a->numHits > b->numHits)
    return 1;
  else if (a->numHits < b->numHits)
    return -1;

  if (a->totalSpace > b->totalSpace)
    return 1;
  else if (a->totalSpace < b->totalSpace)
    return -1;

  return 0;
}

/* Find an id2name string by conversion name ("tag") and id value */
int GetID2NameRecord(FILE *fP, char *tagP,
                     int id, int bufSize, char *bufP)
{
  fmtDesc_t *nextP;

  DFP("GetID2NameRecord(0x%X, 0x%X, 0x%X, 0x%X, 0x%X)\n",
      fP, tagP, id, bufSize, bufP);

  nI2NLookup++;

  /* search hash chain for the specified hookid and tag */
  for (nextP = i2nHashP[I2NHASH(tagP, id)];
       nextP != NULL;
       nextP = nextP->fmtNextP)
  {
    nI2NCheck++;

    /* Check whether hookid and tag match. */
    if (nextP->hookid == id &&
        strncmp(nextP->cidP, tagP, MAX_COMPID_LEN) == 0)
    {
      DFP("GetID2NameRecord for id %X is: %s\n", id, nextP->fmtP);
      strncpy(bufP, nextP->fmtP, bufSize);
      return 0;
    }
  }

  /* not found */
  nI2NFail++;
  return -1;
}

/* Find an id2name string by conversion name ("tag") and id value.
   The id value may be a combination of serveral single values */
int GetID2NameRecords(FILE *fP, char *tagP,
                      UInt64 id, int bufSize, char *bufP)
{
  /* First, see if the id is a single value or not */
  if (0 == GetID2NameRecord(fP, tagP, (int)id, bufSize, bufP))
    return 0;

  /* If not, try to find the hookid list for the id2name tag */
  I2NValueList *curI2NVLP = NULL;
  if (findI2NValueListInHashTable(tagP, &curI2NVLP, 0) != 0)
    return -1;

  /* Found the hookid list for the id2name tag, now extract the single values */
  UInt64 *valuesP = curI2NVLP->valuesP;
  UInt64 myvalues[TRC_MAX_ID2NAME];
  int nValues = curI2NVLP->nValues;

  /* Quick check to see whether the id value is out of range or not */
  if (id < valuesP[nValues-1])
    return -1;

  int i = 0;
  int match = 0;
  for (i = 0; i < nValues; i++)
  {
    if (id >= valuesP[i])
    {
      /* Don't take the value 0 if other values are already found */
      if (valuesP[i] == 0 && match > 0)
        break;

      /* Expect the single values to take a unique bit, or along the same
         lines such that the combination won't be as large as 2 times of
         the largest single value. */
      id -= valuesP[i];
      if (id >= valuesP[i])
        return -1;

      /* found a single value */
      myvalues[match++] = valuesP[i];

      /* Mission accomplished */
      if (id == 0)
        break;

      /* limit the number of single values to find */
      if (match == TRC_MAX_ID2NAME)
        break;
    }
  }

  /* If there is no perfect match, there is no match */
  if (id > 0)
    return -1;

  /* Extract the names of the single id values */
  int used = 0;
  for (i = 0; i < match; i++)
  {
    /* Find the name of each single id value */
    if (0 != GetID2NameRecord(fP, tagP, myvalues[i], bufSize-used, bufP+used))
      return -1;

    /* Make sure the bufP is NULL terminated */
    bufP[bufSize-1] = '\0';

    /* Append a <+> after each new value except the last one */
    used = strlen(bufP);
    if (i < (match - 1) && used < (bufSize - 1))
      bufP[used++] = '+';

    /* The buffer is not large enough for all of the name values */
    if (i < (match - 1) && used >= (bufSize - 1))
      return -1;
  }

  /* Found, so nI2NFail count from the first GetID2NameRecord can be removed */
  nI2NFail--;

  return 0;
}


void ID2Name(FILE* fmt_fp, trc_datahdr_t *recP, i2nSpec_t* i2nP,
             idname_t* id2nameP, argType *argP)
{
  int i, j, id;
  Boolean msgHandler;
  /* Update the argP[] array as necessary to do the substitutions */
  if (i2nP != NULL)
  {
    /* This loop proceses each of the substitution strings */
    for (i = 0; i2nP[i].i2nNameP != NULL; i++)
    {
      int whichSub = i2nP[i].argNum - 1;
      char *argtP  = i2nP[i].i2nNameP;
      char *sP = id2nameP[whichSub].s;

      id = argP[whichSub];

      /* Thread_id2name conversion need special treatment */
      msgHandler = false;
      if (strcmp(argtP, "Thread_id2name") == 0)
      {
        /* Thread IDs >= FirstMsgHandlerThread are really message
           handler IDs, so look for a tscMsg_id2name entry in that case.
           Thread IDs < FirstMsgHandlerThread are "standard" thread IDs,
           so look for StdThread_id2name in that case. */
        if (id >= FirstMsgHandlerThread)
        {
          /* Adjust id value and conversion function.  Set a flag to
             remember to insert "Msg handler" in front of the name that
             is found. */
          id -= FirstMsgHandlerThread;
          argtP = "tscMsg_id2name";
          msgHandler = true;
        }
        else
          argtP = "StdThread_id2name";
      }

      /* Now look-up the substitution for argtP (xx_id2name) that is
         associated with the appropriate traced value (id). */
      if (0 != GetID2NameRecords(fmt_fp, argtP, id, TRC_MAX_ID2NAME, sP))
      {
        /* Unable to find the formatting (substitution) information */

        /* Special notation for NodeAddrs */
        if (0 == strcmp(argtP, "nodeaddr_id2name"))
        {
          unsigned int cl, idx;
          switch (id)
          {
            case -1: sprintf(id2nameP[whichSub].s, "<none>"); break;
            case -2: sprintf(id2nameP[whichSub].s, "<invalid>"); break;
            case -3: sprintf(id2nameP[whichSub].s, "<loopback>"); break;
            default:
              cl = (id >> 24) & 0xff;
              idx = id & 0xffff;
              if (id & 0x10000)
                sprintf(id2nameP[whichSub].s, "<c%dp%d>", cl, idx);
              else
                sprintf(id2nameP[whichSub].s, "<c%dn%d>", cl, idx);
              break;
          }
        }

        /* IpAddrs are traced as four words and printed in either
           IPv4 or IPv6 format. */
        else if (0 == strcmp(argtP, "ipaddr_id2name"))
        {
          union { unsigned int w[4]; unsigned char c[16]; } a;

          a.w[0] = a.w[1] = a.w[2] = a.w[3] = 0;
          for (j = 0;
               j < 4 && whichSub + j < LXTRACE_MAX_FORMAT_SUBS;
               j++)
            a.w[j] = argP[whichSub + j];
          *sP = '\0';

          if (a.w[0] == -1U && a.w[1] == -1U &&
              a.w[2] == -1U && a.w[3] == -1U)
            sprintf(sP, "<none>");
          else if (a.w[0] == 0 && a.w[1] == 0 &&
                   a.c[8] == 0 && a.c[9] == 0 &&
                   a.c[10] == 0xff && a.c[11] == 0xff)
            /* Treat IPv4-mapped IPv6 addresses as IPv4 */
            inet_ntop(AF_INET, &a.c[12], sP, TRC_MAX_ID2NAME);
#ifdef AF_INET6
          //not supported on windows yet
          else
            inet_ntop(AF_INET6, a.c, sP, TRC_MAX_ID2NAME);
#endif
          /* Shift following arguments left */
          for (j = whichSub + 1;
               j < recP->trNArgs && j + 3 < LXTRACE_MAX_FORMAT_SUBS;
               j++)
            argP[j] = argP[j + 3];
        }

        /*  ObjKey is traced as three 64-bit hex numbers */
        else if (0 == strcmp(argtP, "objkey_id2name"))
        {
#if defined( __64BIT__) || defined(_LP64)
          long long *w = argP + whichSub;
          const int shift = 2;
#else
          long long w[3];
          const int shift = 5;
          argType *aP = argP + whichSub;
          w[0] = ((unsigned long long)aP[0] << 32) + (unsigned long)aP[1];
          w[1] = ((unsigned long long)aP[2] << 32) + (unsigned long)aP[3];
          w[2] = ((unsigned long long)aP[4] << 32) + (unsigned long)aP[5];
#endif
          if (w[2] == 0xFFFFFFFFFFFFFFFFLL)
            sprintf(sP, "%08X:%08X:%08X:%08X",
                    High32(w[0]), Low32(w[0]), High32(w[1]), Low32(w[1]));
          else
            sprintf(sP, "%016llX:%016llX:%016llX", w[0], w[1], w[2]);

          /* Shift following arguments left */
          for (j = whichSub + 1;
               j < recP->trNArgs && j + shift < LXTRACE_MAX_FORMAT_SUBS;
               j++)
            argP[j] = argP[j + shift];
        }

        /* lockState to string conversions. */
        else if ((0 == strcmp(argtP, "lockstateHigh_id2name")))
        {
          #define lsAcqPend  0x80000000
          #define lsLxLock   0x40000000
          #define lsXwLock   0x20000000
          #define lsWaLock   0x10000000
          #define lsWwLock   0x08000000
          #define lsRvkLock  0x04000000
          #define lsRoCount  0x03FFF800
          #define lsRoUnit   0x00000800
          #define lsWoCount  0x000007FF
          #define lsWoUnit   0x00000001
          char buf1[32], buf2[32];
          buf1[0]='\0';
          buf2[0]='\0';
          if (id & lsRoCount)
            sprintf(buf1, " ro:%d", (id & lsRoCount)/lsRoUnit);
          if (id & lsWoCount)
            sprintf(buf2, " wo:%d", (id & lsWoCount)/lsWoUnit);
          sprintf(sP, "[%s%s%s%s%s%s%s%s",
                  (id & lsAcqPend) ? " acq" : "",
                  (id & lsLxLock) ? " lx" : "",
                  (id & lsXwLock) ? " xw" : "",
                  (id & lsWaLock) ? " wa" : "",
                  (id & lsWwLock) ? " ww" : "",
                  (id & lsRvkLock) ? " rvk" : "",
                  buf1, buf2);

        }

        /* lockState to string conversions. */
        else if ((0 == strcmp(argtP, "lockstateLow_id2name")))
        {
          #define lsRfCount  0x7FF00000
          #define lsRfUnit   0x00100000
          #define lsWfCount  0x000FFE00
          #define lsWfUnit   0x00000200
          #define lsRsCount  0x000001FF
          #define lsRsUnit   0x00000001
          char buf1[32], buf2[32], buf3[32];
          buf1[0]='\0';
          buf2[0]='\0';
          buf3[0]='\0';
          if (id & lsRfCount)
            sprintf(buf1, " rf:%d", (id & lsRfCount)/lsRfUnit);
          if (id & lsWfCount)
            sprintf(buf2, " wf:%d", (id & lsWfCount)/lsWfUnit);
          if (id & lsRsCount)
            sprintf(buf3, " rs:%d", (id & lsRsCount)/lsRsUnit);
          sprintf(sP, "%s%s%s ]", buf1, buf2, buf3);
        }

        /* lockFlags to string conversions. */
        else if ((0 == strcmp(argtP, "lockflags_id2name")))
        {
          #define lfDaemonWriter      0x0001
          #define lfRevokePending     0x0002
          #define lfRevokeProcessing  0x0004
          #define lfAcqGranted        0x0008
          #define lfWakeupAll         0x0010
          sprintf(sP, "[%s%s%s%s%s ]",
                  (id & lfDaemonWriter)     ? " dmn" : "",
                  (id & lfRevokePending)    ? " rvk" : "",
                  (id & lfRevokeProcessing) ? " pro" : "",
                  (id & lfAcqGranted)       ? " gra" : "",
                  (id & lfWakeupAll)        ? " wka" : "");
        }

        /* Any other case is a lack of proper formatting information
           for the generated trace record.  Insert "unknown" into the
           formatted output and continue. */
        else
        {
#ifdef DBGASSERTS
          DP("%s: %s substitution %d (value=%d) not found (used in hook %08X).\n",
             TRCFMT_NAME, argtP, whichSub+1, argP[whichSub], (recP->trHook&0xFFFFFFFF));
          // To debug one of these, exit here.
          // goto exit_format;
#endif // DBGASSERTS
          sprintf(sP, "unknown(%d)", argP[whichSub]);
        }
      }

      /* If we are converting the thread ID of a message handler,
         insert "Msg handler" in front of the name. */
      if (msgHandler)
      {
        char tmpBuff[TRC_MAX_ID2NAME];
        strncpy(tmpBuff, sP, sizeof(tmpBuff));
        snprintf(sP, TRC_MAX_ID2NAME, "Msg handler %s", tmpBuff);
      }

      /* Alter the TRC_ARGLIST for this substituted parameter */
      argP[whichSub] = (argType)sP;
    } /* for each substitution string */

  } /* End-substitutions */

}
#ifndef GPFS_WINDOWS
/* Convert timestamp to string format. */
extern char *ctime_r(const time_t *, char *);
char *fmtTime(time_t time, char *timeStr)
{
  /* ctime_r returns a null-terminated, readable string that includes
   * a newline char which we remove.  When ctime_r cannot convert the
   * supplied timestamp, it returns NULL.
   */
  if (!ctime_r(&time, timeStr) || time == -1)
    // append space for formatting
    return "unknown                 ";

  timeStr[strlen(timeStr) - 1] = '\0';
  return timeStr;
}

static struct timeval wrap_prev = { 0, 0 };
static struct timeval wrap_next = { 0, 0 };

/* isWrapPoint:
 * See if consecutive timestamps are backwards (indicates a wrap-point).
 */
int isWrapPoint(struct timeval prev, struct timeval next)
{
  /* Detect timestamps going backwards.  This could be the wrap-point
   * or just wobble.
   */
  if ((prev.tv_sec > next.tv_sec) ||
      ((prev.tv_sec == next.tv_sec)
        && (prev.tv_usec - next.tv_usec > allowed_usec_wobble)))
  {
    /* Prepare for simple subtraction (do any necessary borrowing) */
    if (prev.tv_usec < next.tv_usec)
    {
      prev.tv_usec += 1000000;
      prev.tv_sec  -= 1;
    }

    DFP("isWrapPoint: wrap delta %d.%06d (prev %d.06d, next %d.06d)\n",
        (prev.tv_sec - next.tv_sec), (prev.tv_usec - next.tv_usec),
        prev.tv_sec, prev.tv_usec, next.tv_sec, next.tv_usec);

    /* Detect and warn if multiple wrap points are detected. */
    if (((wrap_prev.tv_sec != 0) &&
         ((wrap_prev.tv_sec != prev.tv_sec) || (wrap_prev.tv_usec != prev.tv_usec))) ||
        ((wrap_next.tv_sec != 0) &&
         ((wrap_next.tv_sec != next.tv_sec) || (wrap_next.tv_usec != next.tv_usec))))
    {
      DP("%s", "Multiple wrap-points found:\n");
      DP("\t%d.%06d-%d.%06d (delta %d.%06d)\n",
         wrap_prev.tv_sec, wrap_prev.tv_usec, wrap_next.tv_sec, wrap_next.tv_usec,
         (wrap_prev.tv_sec - wrap_next.tv_sec), (wrap_prev.tv_usec - wrap_next.tv_usec));
      DP("\t%d.%06d-%d.%06d (delta %d.%06d)\n",
         prev.tv_sec, prev.tv_usec, next.tv_sec, next.tv_usec,
         (prev.tv_sec - next.tv_sec), (prev.tv_usec - next.tv_usec));
    }

    /* Keep track of what we have now declared to be a wrap point */
    wrap_prev = prev;
    wrap_next = next;

    return true;
  }

  return false;
}

#ifdef SHARED_TRC_BUFF

/* Find the first and last trace entry for use in emitting the trace stop/start entry */
void findStartStop(FILE *raw_fp, Boolean hasWrapped[])
{
  Int64 wrapPoint = ftell(raw_fp);
  /* We are currently at the beginning of the trcfile, so lets look at the very first entry */
  newHdr_t firstHdr;
  newHdr_t lastHdr;
  long long prevseq = 0;
  struct timeval thisTime;
  int rc = ReadNextTrcHeaderSharedBuff(raw_fp, &firstHdr, &prevseq, hasWrapped);
  /* It's possible for there to be no valid buffers between the wrap point and the
     end of the file. If so, restart at the beginning of the file */
  if (rc == 0)
    {
      fseek(raw_fp, 0, SEEK_SET);
      ReadNextTrcHeaderSharedBuff(raw_fp, &firstHdr, &prevseq, hasWrapped);
    }
  prevseq = 0;
  long long startCycle = firstHdr.h0.trTime;
  /* Now find the last trace entry */
  if (wrapPoint == 0)
    fseek(raw_fp, 0, SEEK_END);
  else
    fseek(raw_fp, wrapPoint-1, SEEK_SET);

  ReadPrevTrcHeaderSharedBuff(raw_fp, &lastHdr, &prevseq);
  long long endCycle = lastHdr.h0.trTime;

  if (tscSpeed == -1)
    thisTime = calculateTimestamp(TODList[0]->cycleCount, TODList[0]->timestamp,
                                  startCycle, TODList[0]->cpuSpeed);
  else
    thisTime = calculateTimestamp(TODList[0]->cycleCount, TODList[0]->timestamp,
                                  startCycle, tscSpeed);
  startTime = thisTime.tv_sec;
  if (tscSpeed == -1)
    thisTime = calculateTimestamp(TODList[currTODCount-1]->cycleCount,
                                  TODList[currTODCount-1]->timestamp, endCycle,
                                  TODList[currTODCount-1]->cpuSpeed);
  else
    thisTime = calculateTimestamp(TODList[currTODCount-1]->cycleCount,
                                  TODList[currTODCount-1]->timestamp, endCycle,
                                  tscSpeed);
  stopTime = thisTime.tv_sec;


  fseek(raw_fp, wrapPoint, SEEK_SET);
}

int hookHasProperPrefix(uint hookid)
{
  uint validHookPrefix[] = { 0x3090,  0x3080,  0x3070,  0x3060 };
  int i;
  uint localHookid;
  localHookid = hookid >> 16;

  for(i = 0; i < 4; i++)
  {
    if( localHookid == validHookPrefix[i])
      return 1;
  }
  return 0;
}

/* This is used only to find the findStartStop to find the last trace entry */
int ReadPrevTrcHeaderSharedBuff(FILE *raw_fp, newHdr_t* hdrP,  long long prevSeq[])
{
    int nRead;
    int bytesRead;
    Int64 currWrapPt;
    newtrc_header_t new_hdr;
    trc_datahdr_t data_hdr;
    int rewoundAlready = 0;

    while ((bytesRead = fread(hdrP, 1, sizeof(newHdr_t), raw_fp)) != 0)
      {
        Int64 prevLocation = ftell(raw_fp) - bytesRead;

        if (hdrP->h0.trMagic == LXTRACE_SWAP_MAGIC)
          {
            int dataLen = hdrP->h0.trLength-sizeof(newHdr_t);
      /* This check should weed out headers with garbage values in them. It's a bit kludgy. */
            if( dataLen < 0 || hdrP->h0.trLength > LXTRACE_MAX_DATA || !hookHasProperPrefix(hdrP->hdr.trHook) ||
                hdrP->hdr.trNArgs < 0)

              {
                DFP("   INVALID. preOffset %d postoffset %d, post+dataLen %d Magic number %d  trTime %lld (0x%lX) trProcess %d trWrapPoint %d, prevWrapPt %d  trCPU %d trLength %d datahdr trHook 0x%lX (%d) trNArgs %d trSPos %d trSLen %d  \n",
                    prevLocation, ftell(raw_fp), ftell(raw_fp) + hdrP->h0.trLength,  (int)hdrP->h0.trMagic, 
                    (Int64)hdrP->h0.trTime, (Int64)hdrP->h0.trTime, (int)hdrP->h0.trProcess, (int)hdrP->h0.trWrapPoint, prevSeq[0],
                     (int)hdrP->h0.trCPU,  (int)hdrP->h0.trLength,(int)hdrP->hdr.trHook,(int)hdrP->hdr.trHook,
                    hdrP->hdr.trNArgs, hdrP->hdr.trSPos, hdrP->hdr.trSLen);
          
                fseek(raw_fp, -(bytesRead)-1, SEEK_CUR);
                continue;
              }
            currWrapPt = (int)hdrP->h0.trWrapPoint;
            if (prevLocation > wrapOffset && wrapNumber == prevSeq[0])
              {
                return -1;
              }
            DFP(" Header. preOffset %lld postoffset %lld, post+dataLen %d Magic number %d trTime %lld (0x%lX) trProcess %d trWrapPoint %d, prevWrapPt %d trCPU %d trLength %d datahdr trHook 0x%lX (%d) trNArgs %d trSPos %d trSLen %d  \n",
                prevLocation, ftell(raw_fp), ftell(raw_fp) + hdrP->h0.trLength,  (int)hdrP->h0.trMagic,
                (Int64)hdrP->h0.trTime, (Int64)hdrP->h0.trTime, (int)hdrP->h0.trProcess, (int)hdrP->h0.trWrapPoint, prevSeq[0],
                (int)hdrP->h0.trCPU,  (int)hdrP->h0.trLength,(int)hdrP->hdr.trHook,(int)hdrP->hdr.trHook,
                hdrP->hdr.trNArgs, hdrP->hdr.trSPos, hdrP->hdr.trSLen);


            prevSeq[0] = currWrapPt;
            break;
          }
        else if (hdrP->h0.trMagic == LXTRACE_TOD_MAGIC)
          {
            fseek(raw_fp, -(bytesRead)-1, SEEK_CUR);
            /* TOD traces are not allowed to wrap the buffer to simplify this case*/
            continue;
          }
        else
          {
            /* Incorrect magic number found */
            fseek(raw_fp, -bytesRead-1, SEEK_CUR);
            if(ftell(raw_fp) == 0)
              {

                if (rewoundAlready)
                  return -1;
                fseek(raw_fp, -sizeof(newHdr_t), SEEK_END);
                rewoundAlready++;
              }
            DFP("Wrong magic number, seek to %d and retry\n",ftell(raw_fp));
          }
      }
      return bytesRead;
}



/* Finds the next trace header for shared trace buffer trcfiles. */
int ReadNextTrcHeaderSharedBuff(FILE *raw_fp, newHdr_t* hdrP,  long long prevSeq[], Boolean hasWrapped[])
{
  int nRead;
  int bytesRead;
  int currWrapPt;
  newtrc_header_t new_hdr;
  trc_datahdr_t data_hdr;
  while ((bytesRead = fread(hdrP, 1, sizeof(newHdr_t), raw_fp)) != 0)
  {
    Int64 prevLocation = ftell(raw_fp) - bytesRead;
    if (hdrP->h0.trMagic == LXTRACE_SWAP_MAGIC)
    {
      int dataLen = hdrP->h0.trLength-sizeof(newHdr_t);
      /* This check should weed out headers with garbage values in them. It's a bit kludgy. */
      if( dataLen < 0 || hdrP->h0.trLength > LXTRACE_MAX_DATA || !hookHasProperPrefix(hdrP->hdr.trHook) ||
          hdrP->hdr.trNArgs < 0)
      { 
        DFP("   INVALID. preOffset %lld postoffset %d, post+dataLen %d Magic number %x  trTime %lld (0x%lX) trProcess %d trWrapPoint %d, prevWrapPt %d trCPU %d trLength %d datahdr trHook 0x%lX (%d) trNArgs %d trSPos %d trSLen %d  \n",
         prevLocation, ftell(raw_fp), ftell(raw_fp) + hdrP->h0.trLength,  (int)hdrP->h0.trMagic,  
         (Int64)hdrP->h0.trTime, (Int64)hdrP->h0.trTime, (int)hdrP->h0.trProcess, (int)hdrP->h0.trWrapPoint, prevSeq[0],
           (int)hdrP->h0.trCPU,  (int)hdrP->h0.trLength,(int)hdrP->hdr.trHook,(int)hdrP->hdr.trHook,
         hdrP->hdr.trNArgs, hdrP->hdr.trSPos, hdrP->hdr.trSLen);

        fseek(raw_fp, 1-(bytesRead), SEEK_CUR);
        continue;
      }
      /* The following if condition is how we recognize that we have
         reached the end of the file if it has wrapped. The problem is that
         sometimes we can have traces that appear after the wrap point.
      */

      currWrapPt = (int)hdrP->h0.trWrapPoint;
      if (prevLocation >= wrapOffset && hasWrapped[0] && currWrapPt == wrapNumber) /* hasWrapped is set to false when we rewind */
      {
        DFP(" Skipping header since it is past the wrap point. preOffset %lld postoffset %lld, post+dataLen %lld Magic number %x trTime %lld (0x%lX) trProcess %d trWrapPoint %d, prevWrapPt %d trCPU %d trLength %d datahdr trHook 0x%lX (%d) trNArgs %d trSPos %d trSLen %d  \n",
         prevLocation, ftell(raw_fp), ftell(raw_fp) + hdrP->h0.trLength,  (int)hdrP->h0.trMagic,  
         (Int64)hdrP->h0.trTime, (Int64)hdrP->h0.trTime, (int)hdrP->h0.trProcess, (int)hdrP->h0.trWrapPoint, prevSeq[0],
           (int)hdrP->h0.trCPU,  (int)hdrP->h0.trLength,(int)hdrP->hdr.trHook,(int)hdrP->hdr.trHook,
         hdrP->hdr.trNArgs, hdrP->hdr.trSPos, hdrP->hdr.trSLen);

        /* We have a trace entry that's past the wrap point and should not be. Skip it this pass */
        continue;
      }
      else if (prevLocation > wrapOffset && !hasWrapped[0] && currWrapPt != wrapNumber) /* hasWrapped is set to false when we rewind */
      {
        DFP(" This is the header we exit on, seeing as we just hit the wrap point (%d): preOffset %lld postoffset %lld, post+dataLen %lld Magic number %x trTime %lld (0x%lX) trProcess %d trWrapPoint %d, prevWrapPt %d trCPU %d trLength %d datahdr trHook 0x%lX (%d) trNArgs %d trSPos %d trSLen %d  \n",
            wrapOffset, 
         prevLocation, ftell(raw_fp), ftell(raw_fp) + hdrP->h0.trLength,  (int)hdrP->h0.trMagic,
         (Int64)hdrP->h0.trTime, (Int64)hdrP->h0.trTime, (int)hdrP->h0.trProcess, (int)hdrP->h0.trWrapPoint, prevSeq[0],
            (int)hdrP->h0.trCPU,  (int)hdrP->h0.trLength,(int)hdrP->hdr.trHook,(int)hdrP->hdr.trHook,
         hdrP->hdr.trNArgs, hdrP->hdr.trSPos, hdrP->hdr.trSLen);

        return -1;
      }

      DFP(" Header. preOffset %lld postoffset %lld, post+dataLen %lld Magic number %X trTime %lld (0x%lX) trProcess %d trWrapPoint %d, prevWrapPt %d trCPU %d trLength %d datahdr trHook 0x%lX (%d) trNArgs %d trSPos %d trSLen %d  \n",
         prevLocation, ftell(raw_fp), ftell(raw_fp) + hdrP->h0.trLength,  (int)hdrP->h0.trMagic,  
         (Int64)hdrP->h0.trTime, (Int64)hdrP->h0.trTime, (int)hdrP->h0.trProcess, (int)hdrP->h0.trWrapPoint, prevSeq[0],
         (int)hdrP->h0.trCPU,  (int)hdrP->h0.trLength,(int)hdrP->hdr.trHook,(int)hdrP->hdr.trHook,
         hdrP->hdr.trNArgs, hdrP->hdr.trSPos, hdrP->hdr.trSLen);

      prevSeq[0] = currWrapPt;
      break;
    }
    else if (hdrP->h0.trMagic == LXTRACE_TOD_MAGIC)
    {
      /* TOD traces are not allowed to wrap the buffer to simplify this case*/
      continue;
    }
    else
    {
      /* Incorrect magic number found */
      fseek(raw_fp, 1-bytesRead, SEEK_CUR);
      DFP("Wrong magic number, seek to %d and retry\n", ftell(raw_fp));
    }
  }
  return bytesRead;
}
#endif /* SHARED_TRC_BUFF */

/* Read the next trace header from the raw trace file.  Normally, this just
   involves the read() for the specified file.  However, it also needs to
   check for wrapping of the file and make the necessary adjustments when we
   reach the wrap point. */
int ReadNextTrcHeader(FILE *raw_fp, trc_hdrs_t* hdrP, int cpu,
                      long long prevSeq[])
{
  int nRead;
  off_t off1, off2;
  off1 = off2 = 0;

  if (0 < (nRead = fread(hdrP, 1, hdrSize[cpu], raw_fp)))
  {
    /* Successful read.  Validate the header. */
    if (hdrP->hdr.trMagic != LXTRACE_MAGIC &&
        hdrP->hdr.trMagic != LXTRACE_SEQ_MAGIC)
    {
      /* Maybe we've reached the wrap point.  See if we can get oriented.
         Backup and try 1 byte further until we find a valid header.  This
         should be the wrap point.  If we can verify this, return the header.
       */
      fseek(raw_fp, 1-nRead, SEEK_CUR);
      off1 = ftell(raw_fp) - 1;
      while (0 < (nRead = fread(hdrP, 1, hdrSize[cpu], raw_fp)))
      {
        /* Keep inching forward looking for a valid device header that
         * includes timestamps that also indicate a wrap ocurred. */
        if (hdrP->hdr.trMagic == LXTRACE_SEQ_MAGIC && prevSeq[cpu] > hdrP->hdr.trSeq ||
            (hdrP->hdr.trMagic == LXTRACE_MAGIC && isWrapPoint(prevTime[cpu], hdrP->hdr.trTime)))
        {
          DFP("ReadNextTrcHeader: wrap point between %d.%d %llu and %d.%d %llu\n",
              hdrP->hdr.trTime.tv_sec, hdrP->hdr.trTime.tv_usec, prevSeq[cpu],
              prevTime[cpu].tv_sec, prevTime[cpu].tv_usec, hdrP->hdr.trSeq);

          prevSeq[cpu] = hdrP->hdr.trSeq;
          /* We've found what appears to be a valid header. */
          prevTime[cpu] = hdrP->hdr.trTime; /* timestamp of the last valid record */
          nWraps += 1;
          DFP("ReadNextTrcHeader: wrap point at %d gap %d\n", off1, off2-off1);

          /* Save the oldest timestamp that we've found */
          if (hdrP->hdr.trTime.tv_sec < startTime)
            startTime = hdrP->hdr.trTime.tv_sec;

          return nRead;
        }
        else if (hdrP->hdr.trMagic == LXTRACE_MAGIC ||
                 hdrP->hdr.trMagic == LXTRACE_SEQ_MAGIC)
        {
          /* We've found a valid header some ways from the beginning of
             the file, but it's not a wrap point.  This could be due to
             some garbage in the file.  Try to skip garbage data. */
          prevSeq[cpu] =hdrP->hdr.trSeq;
          /* timestamp of the last valid record */
          prevTime[cpu] = hdrP->hdr.trTime;

          /* Save the latest timestamp we've found */
          if (hdrP->hdr.trTime.tv_sec > stopTime)
            stopTime = hdrP->hdr.trTime.tv_sec;

          /* Save the oldest timestamp that we've found */
          if (hdrP->hdr.trTime.tv_sec < startTime)
            startTime = hdrP->hdr.trTime.tv_sec;

          return nRead;
        }
        fseek(raw_fp, 1-nRead, SEEK_CUR);
        off2 = ftell(raw_fp);
      }
      DP("%s", "lxtrace format: ReadNextTrcHeader failed to find a valid record!\n");
    }
    else
    {
      /* Valid header found, but this could still be a wrap-point if the
       * alignment worked out exactly.
       */
      if (hdrP->hdr.trMagic == LXTRACE_SEQ_MAGIC && prevSeq[cpu] > hdrP->hdr.trSeq ||
          (hdrP->hdr.trMagic == LXTRACE_MAGIC && isWrapPoint(prevTime[cpu], hdrP->hdr.trTime)))
      {
        DFP("ReadNextTrcHeader: wrap point between %d.%d %llu and %d.%d %llu\n",
            hdrP->hdr.trTime.tv_sec, hdrP->hdr.trTime.tv_usec, prevSeq[cpu],
            prevTime[cpu].tv_sec, prevTime[cpu].tv_usec, hdrP->hdr.trSeq);

        nWraps += 1;
        DFP("ReadNextTrcHeader: wrap point at %d gap %d\n", off1, off2-off1);

        prevSeq[cpu] = hdrP->hdr.trSeq;
        /* timestamp of the last valid record */
        prevTime[cpu] = hdrP->hdr.trTime;

        /* Save the oldest timestamp that we've found */
        if (hdrP->hdr.trTime.tv_sec < startTime)
          startTime = hdrP->hdr.trTime.tv_sec;
      }
      else
      {
        prevSeq[cpu] = hdrP->hdr.trSeq;
        /* timestamp of the last valid record */
        prevTime[cpu] = hdrP->hdr.trTime;

        /* Save the latest timestamp we've found */
        if (hdrP->hdr.trTime.tv_sec > stopTime)
          stopTime = hdrP->hdr.trTime.tv_sec;

        /* Save the oldest timestamp that we've found */
        if (hdrP->hdr.trTime.tv_sec < startTime)
          startTime = hdrP->hdr.trTime.tv_sec;
      }
    }

  }
  return nRead;
}

/* Read the next trace header from the raw trace file.  Normally, this just
   involves the read() for the specified file.  However, it also needs to
   check for wrapping of the file and make the necessary adjustments when we
   reach the wrap point. */
int ReadNextCompHeader(FILE *raw_fp, trc_comp_header_t *hdrP)
{
  int nRead;

  if (0 < (nRead = fread(hdrP, 1, sizeof(trc_comp_header_t), raw_fp)))
  {
    /* Successful read.  Validate the header. */
    if (hdrP->trMagic != LXTRACE_COMP_MAGIC)
    {
      DFP("possible wrap point at %d\n", ftell(raw_fp)-sizeof(trc_comp_header_t));
      /* Maybe we've reached the wrap point.  See if we can get oriented.
         Backup and try 1 byte further until we find a valid header.  This
         should be the wrap point.  If we can verify this, return the header.
       */
      fseek(raw_fp, 1-nRead, SEEK_CUR);
      while (0 < (nRead = fread(hdrP, 1, sizeof(trc_comp_header_t), raw_fp)))
      {
        if (hdrP->trMagic == LXTRACE_COMP_MAGIC )
        {
          DFP("wrap point found at %d\n", ftell(raw_fp)-sizeof(trc_comp_header_t));
          nWraps++;
          return nRead;
        }
        fseek(raw_fp, 1-nRead, SEEK_CUR);
      }
    }
  }
  return nRead;
}

#ifdef SHARED_TRC_BUFF

double calcCPUSpeed(struct refTOD *old, struct refTOD *new)
{
  long long oldCycleCount = old->cycleCount;
  long long newCycleCount = new->cycleCount;
  struct timeval timeDiff = subtrTimeval(new->timestamp, old->timestamp);
  long long microSec = timeDiff.tv_sec * 1000000 + timeDiff.tv_usec;
  long long diffCycleCount = new->cycleCount - old->cycleCount;

  DFP("---->: old %lld new %lld diff %lld. Timestamp %lld.%lld - %lld.%lld = %lld.%lld. diff in microseconds %lld speed %lld\n",
     oldCycleCount, newCycleCount, diffCycleCount, new->timestamp.tv_sec,new->timestamp.tv_usec,
     old->timestamp.tv_sec, old->timestamp.tv_usec, timeDiff.tv_sec, timeDiff.tv_usec, microSec, 
     (microSec ==0)?0:diffCycleCount/microSec);

  /* Need the time difference to be at least 100ms for cpu speed to be determined reliably. */
  if (microSec < 100000 || diffCycleCount == 0 && tscSpeed == -1)
  {
    DFP("---->: time difference is under 100msec to calcualte CPU speed\n");
    return 1;
  }
  if (microSec == 0)  /* dont end up divide by 0 */
    microSec++;
  return diffCycleCount/microSec;
}

void getTODListCPUSpeed()
{
  int i=0;
  int failedCalcCnt=0;
  double cpuSpeed;
  
  /* First pass: use adjacent TODs to determine CPU speed */
  for (i = 1; i < currTODCount; i++)
  {
    if ((cpuSpeed=calcCPUSpeed(TODList[i-1], TODList[i])) != 1)
      TODList[i]->cpuSpeed = cpuSpeed;
  }

  /*Second pass:retry entries that failed in the first pass by using farther TOD */
  for (i = 1; i < currTODCount; i++)
  {
    if (TODList[i]->cpuSpeed ==0)
    {
      /* Try the first and last TOD as reference */
      DFP ("TOD[%d] cycleCount %lld timestamp %lld.%lld failed. Trying first TOD as reference: \n",
             i, TODList[i]->cycleCount, TODList[i]->timestamp.tv_sec, TODList[i]->timestamp.tv_usec);
      if ( (cpuSpeed=calcCPUSpeed(TODList[0], TODList[i])) ==1)
      {
        DFP ("TOD[%d]failed. Trying last TOD as reference\n",i);
        cpuSpeed=calcCPUSpeed(TODList[i], TODList[currTODCount-1]);
      }
      if (cpuSpeed != 1)
        TODList[i]->cpuSpeed = cpuSpeed; 
      else /* track how many failed */
      {
        failedCalcCnt++;
        DFP ("TOD[%d] failed to calucluate CPU speed. failedCalcCnt %d \n", i, failedCalcCnt);
      }
    }
  }

  /* special handling for the first entry */
  if (TODList[1]->cpuSpeed)
     TODList[0]->cpuSpeed = TODList[1]->cpuSpeed;   
  else 
     failedCalcCnt++;

  /* If TOD ends are less than 100ms apart, CPU speed can not be 
      determined. Default to using cycles */
  if (failedCalcCnt == currTODCount)
  {
    DP("Could not calculate approximate timestamps for overwrite mode for any TOD entries failedCalcCnt=%d. Determining cpu speed failed. Falling back to cpu count\n",failedCalcCnt);
    showCycleCountOnly = 1;
    return;
  }

  /* At this point all or some(rare) TODs have CPU speed, fill in 
     the missing ones using the prev entry. Note that the last entry should 
     not be zero as it got a chance to be referenced against the first one */
  for (i = 0; i < currTODCount; i++)
  {
    if (TODList[i]->cpuSpeed ==0)
      TODList[i]->cpuSpeed = (i==0)?TODList[currTODCount -1]->cpuSpeed:TODList[i-1]->cpuSpeed;
      
    DFP ("TOD[%d] cycleCount %lld timestamp %lld.%lld : Using cpu speed from TOD %d\n",
         i, TODList[i]->cycleCount, TODList[i]->timestamp.tv_sec, TODList[i]->timestamp.tv_usec,
         (i==0)?currTODCount -1:i-1);
  } 
  return;
}


/* Find the TOD timestamps so we know (approximately) the time periods
   this trace buffer covers. We seek through the entire file 1 byte
   at a time to do this, then reset the file pointer at the end.
   TODO: To speed this up, either read in larger chunks then parse the output
   Or put the timestamps in the footer so we don't read the file twice. */
void findTODTimestamps(FILE *raw_fp)
{
  newHdr_t hdr;
  newHdr_t *hdrP = &hdr;
  int bytedRead;
  Int64 wrapPoint = ftell(raw_fp);
  time_t curr;
  long long sec, usec;
  int bytesRead;
  int currWrapNum = -1;
  int bufferSizeTOD = 4096;
  char *bufP;
  Int64 offsetInBuf = 0;
  Int64 prevOffset;
  long long endStamp;
  int needsRewind = 1;
  if (wrapNumber == 0)
    needsRewind = 0;
  bufP = (char *)malloc(sizeof(char)*bufferSizeTOD);
  /* Fill the remainder of the buffer. The beginning of the buffer might have some data
     from the previous pass. */

rewoundFile:
  prevOffset = 0;
  while ((bytesRead = fread(bufP + offsetInBuf, 1, bufferSizeTOD - offsetInBuf, raw_fp)) != 0)
  {
    /* Since TOD traces can't wrap I don't need to look at the very end of the buffer */
    for(offsetInBuf = 0; offsetInBuf < bufferSizeTOD - sizeof(newHdr_t) - 3*sizeof(long long); offsetInBuf++)
    {
      if (offsetInBuf + prevOffset > wrapPoint && !needsRewind && wrapNumber != 0)
      {
        /* If we had to wrap, we have already wrapped
           and we get back to the wrap point. */

        DFP("Found wrap point again at %lld\n", offsetInBuf + prevOffset);
        goto doneScanning;
      }
      hdrP = (newHdr_t*)(bufP + offsetInBuf);
  
      if (hdrP->hdr.trHook == LXTRACE_TOD_MAGIC &&
          hdrP->h0.trLength == sizeof(newHdr_t) + 3*sizeof(long long))
      {
        sec = *(long long *)(bufP + offsetInBuf + sizeof(newHdr_t));
        usec = *(long long*)(bufP + offsetInBuf + sizeof(newHdr_t) + sizeof(long long));
        endStamp = *(long long*)(bufP + offsetInBuf + sizeof(newHdr_t) + 2*sizeof(long long));
        if(endStamp != LXTRACE_TOD_END_MAGIC)
        {
          DFP("          found bad timestamp (missing end magic number). Offset %lld: %lld.%lld (cycle %lld), size %d\n",prevOffset +offsetInBuf, sec, usec,
              (Int64)hdrP->h0.trTime, hdrP->h0.trLength);
          continue;
        }
        /* To make sure this TOD entry isn't stale, I need to check for two cases. One after the buffer offset
           where the wrap number for the entry is wrapNumber-1, and the other for before the buffer offset 
           where the wrap number for the entry is wrapNumber */
        if (wrapNumber != 0 && ((needsRewind == 1 && hdrP->h0.trWrapPoint != wrapNumber -1) || (needsRewind == 0 && hdrP->h0.trWrapPoint != wrapNumber)))
            //        if (wrapNumber != 0 && (wrapNumber >  hdrP->h0.trWrapPoint) )
        {
          DFP("          found bad timestamp (wrap number decreases). Offset %lld: %lld.%lld (cycle %lld), size %d wrap %d\n",prevOffset +offsetInBuf, sec, usec,
             (Int64)hdrP->h0.trTime, hdrP->h0.trLength, hdrP->h0.trWrapPoint);
          continue;
        }
        
        DFP("found timestamp. Offset %lld: %lld.%lld (cycle %lld), size %d wrap %d\n",prevOffset +offsetInBuf, sec, usec,
           (Int64)hdrP->h0.trTime, hdrP->h0.trLength, hdrP->h0.trWrapPoint);
        if(sec < startTime)
        {
          startTime = (time_t)sec;
          DFP("  startTime changed to %lld.%lld at cyclecount %lld (location in file %d)\n", startTime, usec, hdrP->h0.trTime,prevOffset +offsetInBuf);
        }
        if(sec > stopTime)
        {
          stopTime = (time_t)sec;
          DFP("  stopTime changed to %lld.%lld at %lld\n", stopTime, usec, hdrP->h0.trTime);
        }
        if (currTODCount < TODMaxListSize)
        {
          TODList[currTODCount] = malloc(sizeof(refTOD));
          TODList[currTODCount]->cycleCount = hdrP->h0.trTime;
          TODList[currTODCount]->timestamp.tv_sec = sec;
          TODList[currTODCount]->timestamp.tv_usec = usec;
          TODList[currTODCount]->cpuSpeed = 0;

          DFP("cycleCount %lld timestamp %lld.%lld speed %f\n",
             TODList[currTODCount]->cycleCount, TODList[currTODCount]->timestamp.tv_sec, TODList[currTODCount]->timestamp.tv_usec,
             TODList[currTODCount]->cpuSpeed);
          currTODCount++;
        }
      }
    }
    /* Move the last bytes of the buffer to the beginning */
    bufP = (char *)memmove(bufP, bufP + offsetInBuf, bufferSizeTOD - offsetInBuf);
    offsetInBuf = bufferSizeTOD - offsetInBuf;
    prevOffset = ftell(raw_fp);
  }
  DFP("Scanned till %d. needsRewind is %d\n", ftell(raw_fp), needsRewind);
  if(needsRewind)
  {
    
    needsRewind = 0;
    fseek(raw_fp, 0, SEEK_SET);
    goto rewoundFile;
  }
  doneScanning:
  fseek(raw_fp, wrapPoint, SEEK_SET);

  /* Need at least two TOD entries to determine CPU spped */
  if (currTODCount >= 2)
    getTODListCPUSpeed();
  free(bufP);
}
#endif

FILE* findWrapPoint(char *rawPrefix, char *rawSuffix, int cpu,
                    int nTrcHdrs[], Boolean hasWrapped[],
                    long long prevSeq[])
{
  char rawTraceFileNameP[PATH_MAX];
  trc_hdrs_t h;
  FILE* raw_fp = NULL;
  int ret;

  if (rawPrefix != NULL)
  {
    if (rawSuffix != NULL)
      sprintf(rawTraceFileNameP, "%s.%s.cpu%d", rawPrefix, rawSuffix, cpu);
    else
      sprintf(rawTraceFileNameP, "%s.cpu%d", rawPrefix, cpu);
  }
  else
    return NULL;
  /* Open the trace file. */
  if (NULL == (raw_fp = fopen(rawTraceFileNameP, "r")))
  {
    return NULL;
  }
  hasWrapped[cpu] = false; /* Signal format processing below we have a wrap file */
  nWraps = 0;         /* ReadNextTrcHeader counts wrap points it encounters.*/
  nTrcHdrs[cpu] = 0;
#ifdef SHARED_TRC_BUFF
  if (writeMode == LXTRACE_NB)
  {
    newHdr_t h;
    int tmp;
    prevSeq[0] = -1;

    fread(&wrapOffset, 1, sizeof(Int64), raw_fp);
    fread(&wrapNumber, 1, sizeof(int), raw_fp);
    DFP("The current wrap number is %d and wrap offset is %lld\n", wrapNumber, wrapOffset);
    fread(&tmp, 1, sizeof(int), raw_fp);
    if (tmp <= 0) /* if the tsc speed is not set for some reason */
      tscSpeed = -1;
    else
      tscSpeed = (double)tmp/1000; /* tscSpeed is reported in khz not mhz */

    if(wrapNumber != 0)
    {
      fseek(raw_fp, wrapOffset, SEEK_SET);
      nWraps++;
      hasWrapped[cpu] = true; /* Signal format procesing below to handle the wrap */
    }
  }
  else
#endif
  {
    while (0 < ReadNextTrcHeader(raw_fp, &h, cpu, prevSeq))
    {
      nTrcHdrs[cpu]++;

      /* If ReadNextTrcHeader detects a wrap, leave and begin formating. */
      if (nWraps)
        break;

      /* skip over trace data */
      fseek(raw_fp, h.hdr.trLength - sizeof(trc_datahdr_t), SEEK_CUR);
    }

    /* If a wrap point was located, back-up so the header can be read again
       as formatting begins with this record.  If the file never wrapped,
       reset to the top of the file and process top-to-bottom. */
    if (nWraps)
    {
      fseek(raw_fp, -hdrSize[cpu], SEEK_CUR);
      hasWrapped[cpu] = true; /* Signal format procesing below to handle the wrap */
    }
    else
      rewind(raw_fp);
  }
  return raw_fp;
}

/* Defect 816587. lxtrace wants a prefix passed in,
   and lxtrace will append the cpuX suffix
   if you pass in the full name of a file, make sure that
   it doesn't end in "cpuX".
   This is done as a preprocessing optimization to prevent
   unnecessary renaming of trcfiles
*/
void checkFileNamePrefix(char *rawPrefix)
{
  DFP("checking the file name prefix for %s\n",rawPrefix);
  struct stat buf;
  int dateStamp, hour, minute, second, noWrapNum;
  char hostName[PATH_MAX];  
  int strlength = strlen(rawPrefix);
  int i = strlength;

  int rc = gethostname(hostName, PATH_MAX);
  if (rc != 0)
    return;

  if (stat(rawPrefix, &buf) != 0) //if the file doesn't exist, just return
    return;
  if (strstr(hostName, "cpu")) //if the hostName has "cpu" in it, return
    return;

  //now check if the end of the string is "cpu%d"
  for (i = strlength -1; rawPrefix[i] != '.'; i--)
  {
    if (!isdigit(rawPrefix[i]))
      break;
    if (i == 0)
      break;
  }
  if (i < 3) //If we might look before the start of the string, abort now
    return;
  if (rawPrefix[i] == 'u' && rawPrefix[i-1] == 'p' && 
      rawPrefix[i-2] == 'c' && rawPrefix[i-3] == '.')
    rawPrefix[i-3] = '\0';
}

//find raw files
int findRawFiles(char *rawPrefix)
{
  char fileName[PATH_MAX];
  int nCPU = 0;
  int rc;
  struct stat buf;

  if (stat(rawPrefix, &buf) == 0)
  {
    if (buf.st_size <= 0)
    {
      errno = ENOENT; /* we use the errno as the return code from lxtrace */
      return 0;
    }
    sprintf(fileName, "%s.cpu0", rawPrefix);
    rename(rawPrefix, fileName);
    return 1;
  }
 
  while (1)
  {
    sprintf(fileName, "%s.cpu%d", rawPrefix, nCPU);
    if (stat((const char*)fileName, &buf) != 0)
      break;
    nCPU++;
  }

  return nCPU;
}

//uncompress raw trace file
FILE *unComp(char *compPrefix, char *compSuffix, int cpu,
             Boolean hasWrapped[])
{
  char compTraceFileNameP[PATH_MAX];
  char uncompFileNameP[PATH_MAX];
  FILE *comp_fp; /* raw trace data file */
  FILE *out_fp; /* output file */
  trc_comp_header_t cmp_hdr;
  unsigned long dlen;
  Boolean compressed;
  void *libz_handle = NULL;
  char *error;

  comp_fp = out_fp = NULL;
  /* work buffers */
  char *src = NULL;   /* compressed  trace data */
  char *dst = NULL;   /* uncompressed trace data */
  int ret;

  sprintf(compTraceFileNameP, "%s.cpu%d", compPrefix, cpu);
  /* Open the trace file. */
  if (NULL == (comp_fp = fopen(compTraceFileNameP, "r")))
  {
    goto exit_uncomp;
  }

  if ( fread(&cmp_hdr, 1, sizeof(trc_comp_header_t), comp_fp) > 0)
  {
    //search wrap point
    if (cmp_hdr.trMagic == LXTRACE_COMP_MAGIC)
    {
      hdrSize[cpu] = sizeof(trc_hdrs_t);
      fseek(comp_fp, 0, SEEK_SET);
      src = malloc(cmp_hdr.dataLength);
      if ( src == NULL )
        goto exit_uncomp;
      dst = malloc(cmp_hdr.dataLength);
      if ( dst == NULL )
        goto exit_uncomp;
      nWraps = 0;
      //read compression record until wrap point found or reach file tail
      while (ReadNextCompHeader(comp_fp, &cmp_hdr)>0)
      {
        DFP("unComp: find record %d at %d\n",
            cmp_hdr.compLength, ftell(comp_fp));
        if (nWraps)
        {
          break;
        }
        fseek(comp_fp, cmp_hdr.compLength, SEEK_CUR);
      }
      //no wrap point found, go back to head
      if (!nWraps)
        rewind(comp_fp);
      //find wrap point, seek to it
      else
      {
        fseek(comp_fp, -sizeof(trc_comp_header_t), SEEK_CUR);
        hasWrapped[cpu] = true;
      }
    }
    else
    {
      //for backward compatibility
      //previous trace header do not have sequence number
      //Also, non-blocking (LXTRACE_NB) traces aren't compressed
      if (cmp_hdr.trMagic == LXTRACE_MAGIC)
        hdrSize[cpu] = sizeof(trc_noseq_header_t) + sizeof(trc_datahdr_t);
      else if (cmp_hdr.trMagic == LXTRACE_SEQ_MAGIC)
        hdrSize[cpu] = sizeof(trc_hdrs_t);
      DFP("unComp: header magic %x header size %d\n",
          cmp_hdr.trMagic, hdrSize[cpu]);
      goto exit_uncomp;
    }

    //open uncompressed raw trace file
    sprintf(uncompFileNameP, "%s.%s.cpu%d", compPrefix, compSuffix, cpu);


    out_fp = fopen(uncompFileNameP, "w+");
    if (out_fp == NULL)
    {
      DP("lxtrace format: could not open file %s, errno %d\n",
         uncompFileNameP, errno);
      goto exit_uncomp;
    }
    //load libz.so.1
    libz_handle = dlopen("libz.so.1", RTLD_NOW);
    if (libz_handle == NULL)
    {
      DP("lxtrace format: dlopen libz.so.1 error: %s\n", dlerror());
      goto exit_uncomp;
    }
    //clear previous dlerror
    dlerror();
    uncompress = dlsym(libz_handle, "uncompress");
    if ((error = dlerror()) != NULL)
    {
      DP("lxtrace format: dlsym uncompress error: %s\n", error);
      goto exit_uncomp;
    }
    //start uncompression from wrap point
    nWraps = 0;
    while (1)
    {
      if (ReadNextCompHeader(comp_fp, &cmp_hdr)>0)
      {
        if (nWraps)
          break;
        ret = fread(src, 1, cmp_hdr.compLength, comp_fp);
        if (ret == cmp_hdr.compLength)
        {
          dlen = cmp_hdr.dataLength;
          ret = uncompress(dst, &dlen, src, cmp_hdr.compLength);
          if ( ret != 0 )
          {
            if (ret == -3)
              DP("%s", "lxtrace format: uncompression data error\n");
            else if (ret == -5)
              DP("%s", "lxtrace format: uncompression buffer error\n");
            else
              DP("%s", "lxtrace format: uncompression error %d\n", ret);
          }
          else if ( dlen != 0 )
          {
            totalCompSize += cmp_hdr.compLength;
            totalPostCompSize += dlen;
            ret = fwrite(dst, dlen, 1, out_fp);
            if (ret == 0)
            {
              DP("lxtrace format: fwrite error %s\n", strerror(errno));
              goto exit_uncomp;
            }
          }
        }
        else
          DP("lxtrace format: compressed file incomplete. Sought %d bytes, get %d bytes\n", cmp_hdr.compLength, ret);
      }
      //go back to file head and uncompress
      else if (hasWrapped[cpu])
      {
        rewind(comp_fp);
        hasWrapped[cpu] = false;
      }
      else break;
    }
  }

exit_uncomp:
  if (libz_handle != NULL)
    dlclose(libz_handle);
  if (src != NULL)
    free(src);
  if (dst != NULL)
    free(dst);
  if (comp_fp != NULL) fclose(comp_fp);
  if (out_fp != NULL)
  {
    rewind(out_fp);
  }
  return out_fp;
}
#ifdef SHARED_TRC_BUFF
/* Obtain the next header; newHdr is a return parameter */
trc_hdrs_t* PickNextTrcHeader(FILE* raw_fp[], int nCPU, trc_hdrs_t cpuHdr[], newHdr_t *newHdr,
                              Boolean hasWrapped[], long long prevSeq[])
#else
trc_hdrs_t* PickNextTrcHeader(FILE* raw_fp[], int nCPU, trc_hdrs_t cpuHdr[],
                              Boolean hasWrapped[], long long prevSeq[])
#endif
{

  int i;
  if (nCPU == 1)
  {
    #ifdef SHARED_TRC_BUFF
    if (writeMode == LXTRACE_NB)
    {
      if (ReadNextTrcHeaderSharedBuff(raw_fp[0], newHdr, prevSeq, hasWrapped)>0)
      {
        if (nWraps ) /* ReadNextTrcHeaderSharedBuff detected and processed a wrap-point */
        {
          DFP("%s", "lxtrace format: ending after processing records 0 to wrap-point (offset %d)\n", ftell(raw_fp[0]));
          nWraps = 0;
          return NULL;
        }
        return (trc_hdrs_t *)newHdr; /* This is needed to be "null" or "not null".
                                        We discard this value in the LXTRACE_NB
                                        case and rely on newHdr as return parameter.*/
      }
      else
        return NULL;
    }
    else
#endif
    {
      if (ReadNextTrcHeader(raw_fp[0], &cpuHdr[0], 0, prevSeq)>0)
      {
        if (nWraps ) // ReadNextTrcHeader detected and processed a wrap-point
        {
          DFP("%s", "lxtrace format: ending after processing records 0 to wrap-point\n");
          nWraps = 0;
          return NULL;
        }
        return &cpuHdr[0];
      }
      else
        return NULL;
    }
  }
#if defined(GPFS_LINUX) && defined(LXTRACE_USE_RELAY)
  //this cpu raw file still have trace data
  if (cpuHdr[curCPU].hdr.trMagic != LXTRACE_NODATA_MAGIC)
  {
    //read trace data from last accessed cpu file. Note that
    //the shared trace buffer case doesn't do per-cpu tracing so it
    //doesn't need to be accounted for
    if (ReadNextTrcHeader(raw_fp[curCPU], &cpuHdr[curCPU], curCPU, prevSeq)>0)
    {
      /* Since we always start AFTER the wrap point, we should only see a
         wrap if we had reached the end of the file and then gone back to
         the beginning to get the records from there up to the wrap point.
         So, if we see this, we have formatted all the records and are done. */

      if (nWraps ) // ReadNextTrcHeader detected and processed a wrap-point
      {
        DFP("%s", "lxtrace format: ending after processing records 0 to wrap-point\n");
        cpuHdr[curCPU].hdr.trMagic = LXTRACE_NODATA_MAGIC;
        nWraps = 0;
      }
    }
    //go back to caller for rewind
    else if (hasWrapped[curCPU])
      return NULL;
    //no data left
    else
      cpuHdr[curCPU].hdr.trMagic = LXTRACE_NODATA_MAGIC;

  }

  //find next cpu file which has data to read
  curCPU = nCPU;
  for (i = 0; i < nCPU; i++)
  {
    if (cpuHdr[i].hdr.trMagic != LXTRACE_NODATA_MAGIC)
    {
      curCPU = i;
      break;
    }
  }
  //all cpu files are consumed
  if (curCPU>=nCPU) return NULL;

  //only last cpu file has data
  if (curCPU == nCPU - 1)
    return &cpuHdr[curCPU];

  //pick up oldest trace record from each cpu trace file
  for (i=curCPU+1; i<nCPU; i++)
  {
    if (cpuHdr[i].hdr.trMagic != LXTRACE_NODATA_MAGIC)
    {
      if (cpuHdr[i].hdr.trSeq < cpuHdr[curCPU].hdr.trSeq)
      {
        curCPU = i;
      }
    }
  }
#endif
  return &cpuHdr[curCPU];
}
#ifdef SHARED_TRC_BUFF
int formatOneTraceRecord(char *compP, char *fmtP, i2nSpec_t *i2nP, char *data,
                         newtrc_header_t *newhdr, trc_hdrs_t *h, trc_datahdr_t *recP,
                         FILE *out_fp, char *fmtString, FILE *fmt_fp,
                         int verbose, int sequence, int useLocal)
#else
int formatOneTraceRecord(char *compP, char *fmtP, i2nSpec_t *i2nP, char *data,
                         trc_hdrs_t *h, trc_datahdr_t *recP,
                         FILE *out_fp, char *fmtString, FILE *fmt_fp,
                         int verbose, int sequence, int useLocal)
#endif
{
  /* Array of pointers to the args for the fprintf that formats the output */
  argType argP[LXTRACE_MAX_FORMAT_SUBS];
  /* work buffers */
  char application_data[LXTRACE_MAX_DATA];   /* raw trace data */
  char timeBufP[26]; /* timestamp in string format */
  int i;
  /* Array of id2name substitutions to be made */
  idname_t id2nameP[LXTRACE_MAX_FORMAT_SUBS];
  int rc;
  double *d;
  memset(application_data, '\0', LXTRACE_MAX_DATA);
  memcpy(application_data, data, LXTRACE_MAX_DATA);
#ifdef SHARED_TRC_BUFF
  double cpuSpeed;
  if (writeMode == LXTRACE_NB)
  {
    if(showCycleCountOnly)
    {
      rc = fprintf(out_fp, fmtString, newhdr->trTime, newhdr->trWrapPoint,
                   newhdr->trProcess, compP);
    }
    else
    {
      if (tscSpeed == -1) /* aka not xLinux */
        cpuSpeed = TODList[currTOD]->cpuSpeed; /* So we can print it out
                                                  at the end of the function. */
      else
        cpuSpeed = tscSpeed;
      struct timeval recordTime = calculateTimestamp(TODList[currTOD]->cycleCount, TODList[currTOD]->timestamp,
                                                     newhdr->trTime, cpuSpeed);
      if (recP->trHook == LXTRACE_TOD_MAGIC)
      {
        if (firstTOD)
          firstTOD--;
        else if (currTOD != currTODCount - 1)
          currTOD++;
      }
      if (tzero.tv_sec == 0 && tzero.tv_usec == 0)
        tzero = recordTime;

      if (recordTime.tv_usec >= tzero.tv_usec)
        recordTime.tv_usec -= tzero.tv_usec;
      else
      {
        recordTime.tv_sec--;
        recordTime.tv_usec += 1000000 - tzero.tv_usec;
      }
      recordTime.tv_sec -= tzero.tv_sec;
      if(showCycleCount)
      {
        //        DP("recordtime %lld.%lld fmtString <%s>\n",recordTime.tv_sec, recordTime.tv_usec, fmtString);
        rc = fprintf(out_fp, fmtString, recordTime.tv_sec, recordTime.tv_usec,
                     newhdr->trProcess, newhdr->trTime, newhdr->trWrapPoint, compP);
      }
      else
      {
        rc = fprintf(out_fp, fmtString, recordTime.tv_sec, recordTime.tv_usec,
                     newhdr->trProcess, compP);
      }
      /* DP(fmtString, recordTime.tv_sec, recordTime.tv_usec, */
      /*    newhdr->trProcess, compP); */
    }
  }
  else
#endif
  {
#ifdef GPFS_SOLARIS
    if (verbose)
      rc = fprintf(out_fp, fmtString,
                   h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                   h->hdr.trThread, h->hdr.trCPU, h->hdr.trBuf, recP->trHook,
                   compP);
    else
      rc = fprintf(out_fp, fmtString,
                   h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                   h->hdr.trThread, compP);
#else
    if (verbose)
    {
      if (sequence)
        rc = fprintf(out_fp, fmtString,
                     h->hdr.trSeq, h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                     h->hdr.trCPU, h->hdr.trBuf, recP->trHook,
                     compP);
      else
        rc = fprintf(out_fp, fmtString,
                     h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                     h->hdr.trCPU, h->hdr.trBuf, recP->trHook,
                     compP);
    }
    else
    {
      if (sequence)
        rc = fprintf(out_fp, fmtString,
                     h->hdr.trSeq, h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                     compP);
      else
      {
        if (useLocal)
          rc = fprintf(out_fp, fmtString,
                       fmtTime(h->hdr.trTime.tv_sec, timeBufP), h->hdr.trTime.tv_usec, h->hdr.trProcess,
                       compP);
        else
          rc = fprintf(out_fp, fmtString,
                       h->hdr.trTime.tv_sec, h->hdr.trTime.tv_usec, h->hdr.trProcess,
                       compP);
      }
    }
#endif
  }
  if (rc < 0)
  {
    DP("lxtrace format: fprintf error %s\n", strerror(errno));
    return 1;
  }

  DFP("raw data: %X %X %X %X %X\n",
      TRC_ARG(0), TRC_ARG(1), TRC_ARG(2), TRC_ARG(3), TRC_ARG(4));
  /* Format-specific processing. */
  if (recP->trSPos == _TR_FORMAT_I)
  {
    /* Trace data consists of n integer values */
    for (i=0; i < recP->trNArgs; i++)
      argP[i] = TRC_ARG(i);
  }
  else if (recP->trSPos < LXTRACE_MAX_FORMAT_SUBS)
  {
 
    /* Trace data contains null terminated string at TRC_ARG(rec.trSPos);
       the length of the string is given by rec.trSLen (rounded up to a
       multiple of ARGLEN).  Get a pointer to the string, make sure it's
       null-terminated (so we don't blow up if the trace data is trash),
       and compute how many extra TRC_ARG words are occupied by the
       string. */
    // ?? should validate trSLen!
    char *s = (char *)&TRC_ARG(recP->trSPos);
    int xi = recP->trSLen/ARGLEN - 1;
    DFP("string at %d; total args %d . application data: 0x%lX, s: 0x%lX, xi: %d, s[%d] = '0', s = %s, argP[0] \n",
       recP->trSPos, recP->trNArgs, application_data, s, xi, recP->trSLen-1, s);
     s[recP->trSLen - 1] = '\0';
     for (i = 0; i < recP->trSPos; i++)
      argP[i] = TRC_ARG(i);
    argP[i++] = (argType)s;
     for (; i <= recP->trNArgs; i++)
      argP[i] = TRC_ARG(i + xi);

  }
  else if (recP->trSPos == _TR_FORMAT_F)
  {
    /* First n words are integer values; a double value is stored at
       TRC_ARG(rec.trNArgs) */
    d = (double *)&TRC_ARG(recP->trNArgs);

    for (i = 0; i < recP->trNArgs; i++)
      argP[i] = TRC_ARG(i);
  }
  else if (recP->trSPos == _TR_FORMAT_X)
  {

    /* XTrace type (preformatted) record. */

    fputs(application_data, out_fp);
    i = strlen(application_data);
    if (i == 0 || application_data[i-1] != '\n')
      putc('\n', out_fp);
    return 0;
  }
  else
  {
    DP("lxtrace format: format %d not supported\n", recP->trSPos);
    fprintf(out_fp, "Unsupported format %d\n", recP->trSPos);
    return 1;
  }
  /* Perform id2name substitutions */
  /* Defect 801461, always do this check instead of having it under DEBUGASSERTS*/
  if (i2nP == NULL &&
      (recP->trSPos==_TR_FORMAT_I || recP->trSPos==_TR_FORMAT_F) &&
      strstr(fmtP, "%s"))
  {
    DP("lxtrace format: hook %X is %s format record containing a string specification.\n",
       (recP->trHook&0xFFFFFFFF),
       (recP->trSPos==_TR_FORMAT_I) ? (char *) "an integer" :
       (recP->trSPos==_TR_FORMAT_F)? (char *) "a float":
       (char *) "an unknown");
    fprintf(out_fp, "lxtrace format: hook %X is %s format record with %d integer args containing a string specification.\n",
            (recP->trHook&0xFFFFFFFF),
            (recP->trSPos==_TR_FORMAT_I) ? (char *) "an integer" :
            (recP->trSPos==_TR_FORMAT_F)? (char *) "a float":
            (char *) "an unknown",
            recP->trNArgs);
    //print to both the outfile and stderr
    DP("%s", "               The formatting information for this record does not provide any substitutions.\n");
    fprintf(out_fp, "%s", "The formatting information for this record does not provide any substitutions:\n");
    fprintf(out_fp, "%s\n", fmtP);
    return 0;
  }
  ID2Name(fmt_fp, recP, i2nP, id2nameP, argP);
  if (recP->trSPos == _TR_FORMAT_F)
  {
    DFP("Double %f, case %i\n", *d, recP->trNArgs);
    switch(recP->trNArgs)
    {
       case 1: rc = fprintf(out_fp, fmtP, TRC_ARGLIST1F); break;
       case 2: rc = fprintf(out_fp, fmtP, TRC_ARGLIST2F); break;
       case 3: rc = fprintf(out_fp, fmtP, TRC_ARGLIST3F); break;
       case 4: rc = fprintf(out_fp, fmtP, TRC_ARGLIST4F); break;
       case 5: rc = fprintf(out_fp, fmtP, TRC_ARGLIST5F); break;
       case 6: rc = fprintf(out_fp, fmtP, TRC_ARGLIST6F); break;
       case 7: rc = fprintf(out_fp, fmtP, TRC_ARGLIST7F); break;
       case 8: rc = fprintf(out_fp, fmtP, TRC_ARGLIST8F); break;
       case 9: rc = fprintf(out_fp, fmtP, TRC_ARGLIST9F); break;
       case 10: rc = fprintf(out_fp, fmtP, TRC_ARGLIST10F); break;
    }
  }
  else
    rc = fprintf(out_fp, fmtP, TRC_ARGLIST);

//  DFP("Invoked %d, errno %d <%s>\n", rc, errno, strerror(errno));
  /* end processing this trace record */
  return 0;
}

/* Print command syntax and exit */
void Usage()
{
  DP("%s", "\n\
Usage: lxtrace { on on-args | off | format format-args | dump | fsync }\n\
\n\
  Command line format for lxtrace on:\n\
    lxtrace on [trcFile] [-r] [-p] [-w] [-s filesize] [-b bufsize] [-n bufferNum]\n\
               [-i interval] [-c complevel] [-l logFile]\n\
      -r         Turn on relayfs trace\n\
      -w         Turn on overwrite mode\n\
      -p         Turn on per-cpu buffer trace\n\
      -t         Turn on coarse-grained timestamp\n\
      -k         Rename raw trace file upon file wrap to keep all trace data\n\
      trcFile    Name of raw trace file to be created.  Default is\n\
                   /tmp/lxtrace.trc.\n\
      filesize   Optional maximum size of wrap-around trace file.  A value\n\
                   of 0 means that the trace file grows indefinitely.  Default\n\
                   is 16M.\n\
      bufsize    Optional size of each buffer.  Default is 4M.\n\
      bufferNum  Optional subbuffer number fora each cpu.  Default is 4.\n\
      interval   Optional trace flush interval.  Default is 15.\n\
      complevel  Optional trace compression level.  Default is 6.\n\
      logFile    Optional file to log all lxtrace daemon activity.\n\
                 Default is to stdout/stderr\n\
\n\
  Command line format for lxtrace format:\n\
    lxtrace format [-T | -D] [-v] [-t formatFile] [-w wobble] [-o outFile] [trcFile]\n\
      -T         Display timestamps as seconds since 1/1/1970.  Default\n\
                   is to display the first timestamp as 0.0 and other\n\
                   timestamps as elapsed time since the first trace record.\n\
      -D         Display timestamps in local calendar format. This option \n\
                   can not be used together with -T option.\n\
      -v         Verbose mode.  Include CPU number, buffer number, and hook\n\
                  word in trace output.  Default is to not include these.\n\
      wobble     Number of microseconds wobble allowed in timestamps.\n\
                 An apparent wrap within the wobble limit will be ignored.\n\
      formatFile Name of file containing trace formats.  Default is\n\
                   /usr/lpp/mmfs/mmfs.trcfmt.\n\
      outFile    Name of file to receive formatted trace output.  Default\n\
                   is standard output.\n\
      trcFile    Name of input raw trace file.  Default is /tmp/lxtrace.trc.\n\
      --stat     Generate statistics about the trcfile being formatted\n"
     );
  if (TraceHandle != -1)
    close(TraceHandle);
  exit(1);
}

/* The lxtrace command.  */
#define LXTRACE_DEBUG_ENV  "LXTRACE_DEBUG"
int main(int argc, char *argv[])
{
  int rc = 0;
  struct kArgs args; // command arguments
  int i;
  int nTrcHdrs[LXTRACE_NR_CPUS];
  Boolean hasWrapped[LXTRACE_NR_CPUS] = { 0 };

  if (argc < 2)
    Usage();

   /* check if LXTRACE_DEBUG environment variable is set to enable DFP debugs */
   if ((getenv(LXTRACE_DEBUG_ENV) != NULL) && atoi(getenv(LXTRACE_DEBUG_ENV)) == 1)
     lxtraceDebug = true;

  /* A new trace is being started.  Validate the options, open the files,
     start the daemon and configure the device. */
  if (strcmp(argv[1], "on") == 0)
  {
    int hw, dpid;
    struct sigaction action;
    char* trcFileNameP = NULL;
    trc_msg msg;
    int mode;
    char *daemonLogFile = NULL;
    /* Initialize local state fields. */

    bufP = NULL;
    outSize = DEF_TRC_FILESIZE;
    bufNum = DEF_TRC_BUFNUM;
    interval = DEF_TRC_INT;
    compLevel = DEF_TRC_COMPLEVEL;
    currentPos = 0;
    nWraps = 0;

    /* Parse arguments */
    for (i = 2; i < argc; i++)
    {
      if (strcmp(argv[i], "-r") == 0)
        useRelay = 1;
      else if (strcmp(argv[i], "-p") == 0)
        relayBuf |= RL_BUF_PERCPU;
      else if (strcmp(argv[i], "-t") == 0)
        relayBuf |= RL_BUF_CTIME;
      else if (strcmp(argv[i], "-w") == 0)
        writeMode = LXTRACE_NB;
      else if (strcmp(argv[i], "-k") == 0)
        noWrap = 1;
      else if (strcmp(argv[i], "-s") == 0  &&  i+1<argc)
      {
        i += 1;
        /* Save the output filesize.  If the size is explicitly given as zero,
           do not wrap. This parameter is not used in overwrite mode. */
        outSize = atoll(argv[i]);
      }
      else if (strcmp(argv[i], "-b") == 0  &&  i+1<argc)
      {
        i += 1;
        #ifdef SHARED_TRC_BUFF
        /* If specified, set the trace device buffer size (otherwise use default) */
        bufSize = atoll(argv[i]) & 0xfffffffffffff000;
        #else
        bufSize = atoi(argv[i]) & 0xfffff000;
#endif /* SHARED_TRC_BUFF */
      }
      else if (strcmp(argv[i], "-n") == 0  &&  i+1<argc)
      {
        i += 1;
        bufNum = atoi(argv[i]);
        if (bufNum < MIN_TRC_BUFNUM || bufNum > MAX_TRC_BUFNUM)
        {
          DP("Buffer number must be between %d and %d\n",
             MIN_TRC_BUFNUM, MAX_TRC_BUFNUM);
          Usage();
        }
      }
      else if (strcmp(argv[i], "-i") == 0  &&  i+1<argc)
      {
        i += 1;
        interval = atoi(argv[i]);
        if (interval < -1)
        {
          DP("Interval must be more than %d\n", -1);
          Usage();
        }
      }
      else if (strcmp(argv[i], "-c") == 0  &&  i+1<argc)
      {
        i += 1;
        compLevel = atoi(argv[i]);
        if (compLevel < MIN_TRC_COMPLEVEL || compLevel > MAX_TRC_COMPLEVEL)
        {
          DP("Compression level must be between %d and %d\n",
             MIN_TRC_COMPLEVEL, MAX_TRC_COMPLEVEL);
          Usage();
        }
      }
      else if (strcmp(argv[i], "-l") == 0  &&  i+1<argc)
      {
        i += 1;
        daemonLogFile = strdup(argv[i]);
        if (daemonLogFile == NULL)
        {
          DP("E_NOMEM in strdup for -l\n");
          exit(1);
        }
      }

      else
        trcFileNameP = argv[i];
    }

    /* Get output file name */
    if (trcFileNameP == NULL )
      trcFileNameP = "/tmp/lxtrace.trc";
    strcpy(percpuOutBasename, trcFileNameP);

    //set default buffer size
    if (!bufSize)
    {
      if ( writeMode == LXTRACE_NB )
        bufSize = DEF_TRC_OVERWRITE_BUFSIZE & 0xfffff000;
      else
        bufSize = DEF_TRC_BUFSIZE & 0xfffff000;
    }

#ifdef SHARED_TRC_BUFF
  TraceHandle = open(TRC_DEVICE, O_RDWR);
  if (TraceHandle < 0)
  {
    /* Device does not exist, wrong type, or already open, ... */
    DP("trace daemon: device open failed with errno %d\n", errno);
    rc = ENODEV;
    return rc;
  }
  if (writeMode == LXTRACE_NB)
  {
    /* Use tsctl to let the daemon know that tracing is going on. */
    int *trcBufferP = NULL;
    struct kxArgs
    {
      int * arg1;
      int * arg2;
      int * arg3;
      int * arg4;
      int * arg5;
      int * arg6;
    };
    struct kxArgs args;
    /* We use bufSize and not outSize in overwrite mode. */
    args.arg1 = (int *) &bufSize;
    DFP("Calling trc_alloc ioctl with arg1 %d (0x%lX) \n",
        *(args.arg1), args.arg1);
    int rc = ioctl(TraceHandle, trc_alloc, &args);
    if(rc != 0)
    {
      DP("trc_alloc ioctl failed, rc %d!\n", rc);
      exit(1);
    }
    else
      DFP("ioctl succeeds\n");
    goto exit;
  }
#endif

    /* Ignore trace file size limit for overwrite mode. */
    if ( writeMode == LXTRACE_NB )
      outSize = 0;
    /* The output filesize must be at least as big as the selected buffer size. */
    if (((outSize < MIN_TRC_FILESIZE) || (outSize > MAX_TRC_FILESIZE)) &&
        (outSize != 0))
    {
      DP("Output file size must be 0 or between %d and %lld\n",
         MIN_TRC_FILESIZE, MAX_TRC_FILESIZE);
      Usage();
    }

    if ((outSize != 0) && (outSize < bufSize))
    {
      DP("lxtrace: the filesize (%d) must be at least as large as the bufsize (%d)\n",
         outSize, bufSize);
      Usage();
    }

    if ( writeMode == LXTRACE_NB )
    {
      /* No compression in overwrite mode */
      compLevel = 0;
      if (bufSize < MIN_TRC_OVERWRITE_BUFSIZE)
      {
        DP("Overwrite buffer size must be more than %d\n",
           MIN_TRC_OVERWRITE_BUFSIZE);
        Usage();
      }
    }
    else
    {
      if (bufSize < MIN_TRC_BUFSIZE)
      {
        DP("lxtrace: trace buffer has size %d instead of the requested %d\n",
           MIN_TRC_BUFSIZE, bufSize);
        bufSize = MIN_TRC_BUFSIZE;
      }
      else if (bufSize > MAX_TRC_BUFSIZE)
      {
        DP("lxtrace: trace buffer has size %d instead of the requested %d\n",
           MAX_TRC_BUFSIZE, bufSize);
        bufSize = MAX_TRC_BUFSIZE;
      }
    }

    //for recycle and daemon error message
    msgQueue = msgGet(IPC_CREAT | 0666);
    if (msgQueue == -1)
    {
      DP("trace daemon: msgget error %d\n", errno);
      rc = errno;
      goto exit;
    }

    /* The trace daemon does all the work.  It is created here ("lxtrace on") and
       lives until "trace off" sends it a SIGTERM signal. */
    dpid = fork();

    if (dpid < 0)
    {
      DP("%s", "lxtrace on: unable to fork trace daemon");
      rc = EAGAIN;
      goto exit;
    }

    else if (dpid != 0)
    {
      /* Parent */
      //wait for child daemon initialization message
      for (i = 0; i < 100; i++)
      {
        //do not wait in case the child exit abnormally and the parent will be blocked forever
        rc = msgrcv(msgQueue, &msg, LXTRACE_MSG_LEN, LXTRACE_MTYPE, IPC_NOWAIT);
        if (rc == -1)
        {
          rc = errno;
          //no message, sleep for 1/10th second and retry
          if (errno == ENOMSG )
          {
            usleep(100000);
          }
          else
          {
            DP("lxtrace on: msgrcv error %d\n", errno);
            break;
          }
        }
        else
        {
          rc = 0;
          break;
        }
      }
      DFP("lxtrace on: receive message from daemon: %s\n", msg.mtext);
      msgctl(msgQueue, IPC_RMID, NULL);
      //return initialization result to mmtrace
      if (rc != 0 || strcmp(msg.mtext, LXTRACE_INIT_ERR) == 0)
      {
        return -1;
      }
      else
        return 0;
    }

    else
    { /* Child/Daemon */

      if (useRelay)
      {
        if (daemonLogFile != NULL)
        {
          int fd = open(daemonLogFile, O_CREAT|O_RDWR|O_SYNC|O_TRUNC, 0644);
          if (fd <= 0)
          {
            //even if the daemonLogFile code fails, still run RelayStart
            fprintf(stdout, "open(%s) failed, rc %d errno %d (%s)\n",
                    daemonLogFile, fd, errno, strerror(errno));
          }
          else
          {
            DFP("open(%s) for daemon log succeeded, using fd %d\n", daemonLogFile, fd);
            int rc = dup2(fd, 1); //reroute stdout to fd
            if (rc != 0)
              DFP("dup2 failed with rc %d, errno %d (%s\n",
                  rc, errno, strerror(errno));
            rc = dup2(fd, 2); // reroute stderr to fd
            if (rc != 0)
              DFP("dup2 failed with rc %d, errno %d (%s\n",
                  rc, errno, strerror(errno));
          }
        }

#if defined(GPFS_LINUX) && defined(LXTRACE_USE_RELAY)
        rc = RelayStart(outSize, bufSize, bufNum, interval, compLevel);
#endif
        goto exit_on;
      }

    } /* child/daemon */
exit_on:
    if (msgQueue != -1 )
    {
      if (rc!=0)
      {
        //initialization failure, send message to parent
        //the parent will remove the message queue
        msg.mtype = LXTRACE_MTYPE;
        strcpy(msg.mtext, LXTRACE_INIT_ERR);
        msgsnd(msgQueue, &msg, strlen(msg.mtext)+1, 0);
        DFP("trace daemon: send message to parent: %s\n", msg.mtext);
      }
    }
    /* exit here for failures during daemon initialization. */
    DFP("%s", "trace daemon exiting\n");
    if (TraceHandle > 0)
      close(TraceHandle);
  }


  /* Format the trace file. */
  else if (strcmp(argv[1], "format") == 0)
  {
    FILE *raw_fp[LXTRACE_NR_CPUS]={ NULL }; /* raw trace data file */
    FILE *fmt_fp; /* the application formatting file */
    FILE *out_fp; /* output file */
    FILE *stat_fp; /* output file for statistics */
    trc_hdrs_t *h;
    int useTOD = 0;    /* true if -T option given */
    int useLocal = 0;  /* true if -D option given */
    int verbose = 0;   /* true if -v option given */
    int sequence = 0;   /* true if -s option given */
    char* formatFileNameP = "/usr/lpp/mmfs/mmfs.trcfmt";
    char* outputFileNameP = NULL;
    char *statFileNameP = NULL;
    char* rawTraceFileNameP = "/tmp/lxtrace.trc";
    char* compSuffix = "uncomp";
    char *compP, *fmtP;
    i2nSpec_t *i2nP;
    int nRecordsProcessed = 0;  /* useful when running under gdb */
    int i, j, id;
    Boolean msgHandler;
    char timeBufP[26]; /* timestamp in string format */
    FILE *version_fp;
    trc_datahdr_t *recP;
    trc_hdrs_t cpuHdr[LXTRACE_NR_CPUS];

#ifdef SHARED_TRC_BUFF
    newHdr_t newHdr;
    TODList = (refTOD **)malloc(sizeof(refTOD)*TODMaxListSize);
    if(TODList == NULL)
    {
      rc = ENOMEM;
      goto exit_format;
    }
    memset(TODList, '0', sizeof(refTOD)*TODMaxListSize);
    currTOD = 0;
#endif
    unsigned long long prevSeq[LXTRACE_NR_CPUS];
    int prevWrapPt;
    /* work buffers */
    char application_data[LXTRACE_MAX_DATA];   /* raw trace data */

    /* Array of id2name substitutions to be made */
    idname_t id2nameP[LXTRACE_MAX_FORMAT_SUBS];
    char fmtString[256];

    /* Array of pointers to the args for the fprintf that formats the output */
    argType argP[LXTRACE_MAX_FORMAT_SUBS];

    int nCPU = 1;
    int fileWrapped = 0;
    fmt_fp = out_fp = stat_fp = NULL;
    
    formatFileP = NULL;
    fmtDescSpaceP = NULL;

    unsigned long long lastSeq = 0;
    unsigned long long lostRecords = 0;
#if (defined(__64BIT__) || defined(_LP64))
    startTime = 9223372036854775807;
#else
    startTime = 2147483647;
#endif
    stopTime = 0;

    /* Parse arguments */
    for (i = 2; i < argc; i++)
    {
      if (strcmp(argv[i], "-T") == 0)
        useTOD = 1;
      if (strcmp(argv[i], "-D") == 0)
        useLocal = 1;
#ifdef SHARED_TRC_BUFF
      else if (strcmp(argv[i], "--showCycleCount") == 0)
        showCycleCount = 1;
#endif
      else if (strcmp(argv[i], "-v") == 0)
        verbose = 1;
      else if (strcmp(argv[i], "-s") == 0)
        sequence = 1;
      else if (strcmp(argv[i], "--stat") == 0)
        genStats = 1;
      else if (strcmp(argv[i], "--statfile") == 0)
        {
          i += 1;
          statFileNameP = argv[i];
        }
      else if (strcmp(argv[i], "-w") == 0  &&  i+1<argc)
      {
        char *endptr;
        i += 1;
        allowed_usec_wobble = strtol(argv[i], &endptr, 10);
        if ((*endptr != 0) || (allowed_usec_wobble < 0))
        {
          DP("%s", "lxtrace format: -w requires an integer argument greater than or equal to zero.\n");
          rc = 12;
          goto exit_format;
        }
      }
      else if (strcmp(argv[i], "-t") == 0  &&  i+1<argc)
      {
        i += 1;
        formatFileNameP = argv[i];
      }
      else if (strcmp(argv[i], "-o") == 0  &&  i+1<argc)
      {
        i += 1;
        outputFileNameP = argv[i];
      }
      else
        rawTraceFileNameP = argv[i];
    }

    /* Open the formatting file. */
    if (NULL == (fmt_fp = fopen(formatFileNameP, "r")))
    {
      DP("lxtrace format: could not open file %s, errno %d\n", formatFileNameP, errno);
      rc = errno;
      goto exit_format;
    }

    if (0 != (rc = BuildFormatTable(fmt_fp)))
    {
      DP("lxtrace format: could not parse file: %s errno %d\n",
         formatFileNameP, errno);
      rc = errno;
      goto exit_format;
    }

    /* Open and truncate the output file. (defaults to stdout) */
    if (outputFileNameP == NULL)
      out_fp = stdout;
    else
    {
      out_fp = fopen(outputFileNameP, "w");
      if (out_fp == NULL)
      {
        DP("lxtrace format: could not open file %s, errno %d\n",
           outputFileNameP, errno);
        rc = errno;
        goto exit_format;
      }
    }
    /* Open and truncate the stat file. (defaults to stdout) */
    if (statFileNameP == NULL)
      stat_fp = stdout;
    else
    {
      stat_fp = fopen(statFileNameP, "w");
      if (stat_fp == NULL)
      {
        DP("lxtrace format: could not open file %s, errno %d\n",
           statFileNameP, errno);
        rc = errno;
        goto exit_format;
      }
    }

    //the code wants this to be a file prefix, not the full filename
    //make sure this is the case
    checkFileNamePrefix(rawTraceFileNameP);

    nCPU = findRawFiles(rawTraceFileNameP);
    if ( nCPU == 0)
    {
      DP("lxtrace format: could not find raw file %s, errno %d\n",
         rawTraceFileNameP, errno);
      rc = errno;
      goto exit_format;
    }
    #ifdef SHARED_TRC_BUFF
    writeMode = checkTraceFileMode(rawTraceFileNameP);
    if (writeMode == LXTRACE_NB)
    {
      DFP("Detected tracefile using overwrite mode\n");
    }
    else
      DFP("Detected tracefile using blocking mode\n");
    #endif

    /* The trace file (raw_fp) is a wrap-around file.  We need to read it to
       see if it ever wrapped.  If it never wrapped, rewind and process it
       top-to-bottom.  If the file did wrap, set the file pointer to the oldest
       record and begin processing from there.  ReadNextTrcHeader detects the
       wrap point.  When formatting of a wrapped file reaches end-of-file, it
       must then go back and read starting at the beginning and stop when
       the wrap point is again reached. */
    for (i=0; i<nCPU; i++)
    {
      //uncompress and find wrap point
      raw_fp[i] = unComp(rawTraceFileNameP, compSuffix, i, hasWrapped);
      if (raw_fp[i] == NULL)
      {
        #ifdef SHARED_TRC_BUFF
        if (writeMode == LXTRACE_NB)
          hdrSize[0] = sizeof(newHdr_t);
        #endif
        /* no compression, find wrap point directly */
        raw_fp[i] = findWrapPoint(rawTraceFileNameP, NULL, i,
                                  nTrcHdrs, hasWrapped, prevSeq);
      }
      else
      {
        /* Compression was used, so we need to scan through for the
           start/end times. During unComp(), the resulting trcfile.uncomp
           files should have had the wrap point placed at the beginning
           of the file. This scan is just to find the timestamps. */
        Boolean hasWrapped_tmp[LXTRACE_NR_CPUS]; //just a placeholder
        (void) findWrapPoint(rawTraceFileNameP, compSuffix, i,
                             nTrcHdrs, hasWrapped_tmp, prevSeq);
      }
      if (raw_fp[i] == NULL)
      {
        rc = 1;
        goto exit_format;
      }
      /* Show the point we interpreted as the wrap-point */
      if (nWraps)
      {
        DFP("CPU %d Wrap point  : %d.%06d-%d.%06d (delta %d.%06d)\n", i,
            wrap_prev.tv_sec, wrap_prev.tv_usec, wrap_next.tv_sec, wrap_next.tv_usec,
            (wrap_prev.tv_sec-wrap_next.tv_sec), (wrap_prev.tv_usec-wrap_next.tv_usec));
      }
      DFP("wrap analysis: nTrcHdrs %d nWraps %d seekpos %d\n",
          nTrcHdrs[i], nWraps, ftell(raw_fp[i]) - hdrSize[i]);
      #ifdef SHARED_TRC_BUFF
      if (writeMode == LXTRACE_NB)
      {
        findTODTimestamps(raw_fp[i]);

        if (currTODCount > 1 || (currTODCount == 1 && tscSpeed != -1))
          findStartStop(raw_fp[i], hasWrapped);
        else
        {
          DP("We found %d TOD entries, which is not enough to convert them to timestamps\n", currTODCount);
          showCycleCountOnly = 1;
          /* Report "unknown" for the time */
          stopTime = -1;
          startTime = -1;
        }
      }
#endif
    }

    /* Begin printing out header information */
    fprintf(out_fp, FMT_TRC_BEGIN, fmtTime(startTime, timeBufP));
    fprintf(out_fp, FMT_TRC_END, fmtTime(stopTime, timeBufP),
            fileWrapped? FILE_WRAP_STR:FILE_NOWRAP_STR);
    /* Achieved compression ratio. Not used for overwrite mode */

    if (writeMode != LXTRACE_NB)
    {
      if (totalCompSize != 0)
        fprintf(out_fp, FMT_COMP_RATIO, totalPostCompSize / totalCompSize);
      else
        fprintf(out_fp, FMT_COMP_RATIO, 0);
    }
    /* Write kernel version to the trace file.  This is the machine
     * where the format was done, but this would normally be the
     * same machine that generated the raw trace as well.
     */
#ifdef GPFS_SOLARIS
    version_fp = fopen("/etc/nodename", "r");
#else
    version_fp = fopen("/proc/version", "r");
#endif
    if (version_fp)
    {
      int verLen;
      const int verStrLen = 512;
      char version_str[verStrLen];

      memset(version_str, 0, verStrLen);
      if (0 < (verLen = fread(version_str, 1, verStrLen, version_fp)))
      {
        version_str[verStrLen-1] = '\0';
        fprintf(out_fp, "Formatted on: %s\n\n", version_str);
      }

      fclose(version_fp);
    }

    /* Reset to avoid reporting a wrap on the first record from the next
     * call to ReadNextTrcHeader.
     */
    for (i=0; i<nCPU; i++)
    {
      prevTime[i].tv_sec = 0;
      prevTime[i].tv_usec = 0;
      prevSeq[i] = 0;
    }
    /* Print first line of header and build trace format string */
    fmtString[0] = '\0';
#ifdef SHARED_TRC_BUFF
    if (writeMode == LXTRACE_NB)
    {
      DFP("Setting up for shared trace overwrite mode\n");
      char hdrString1[256] = {0};
      char hdrString2[256] = {0};

      if(showCycleCountOnly)
      {
        strcat(fmtString, "%17lld %5d %6d %s: ");
        strcat(hdrString1, " CPU Cycle Count Wrap    Pid  COMPONENT_TAG:  application trace record\n");
        strcat(hdrString2, "---------------- ----- -------- ------ --------------------------\n");
      }
      else
      {
        if(useTOD)
        {
          strcat(hdrString1, H1_TS_T);
          strcat(hdrString2, H2_TS_T);
          strcat(fmtString, FMT_TS_T);
        }
        else
        {
          strcat(hdrString1, H1_TS);
          strcat(hdrString2, H2_TS);
          strcat(fmtString, FMT_TS);
        }
        strcat(hdrString1, H1_PID);
        strcat(hdrString2, H2_PID);
        strcat(fmtString, FMT_PID);

        if(showCycleCount)
        {
          strcat(hdrString1, H1_CYCLE);
          strcat(hdrString2, H2_CYCLE);
          strcat(fmtString, FMT_CYCLE);
        }

        strcat(hdrString1, H1_TAG);
        strcat(hdrString2, H2_TAG);
        strcat(fmtString, FMT_TAG);
      }
      fprintf(out_fp, hdrString1);
      fprintf(out_fp, hdrString2);
    }
    else
#endif
    {
      if (sequence)
      {
        fprintf(out_fp, H1_SEQ);
        strcat(fmtString, FMT_SEQ);
      }
      if (useTOD)
      {
        fprintf(out_fp, H1_TS_T);
        strcat(fmtString, FMT_TS_T);
      }
      else if (useLocal)
      {
        fprintf(out_fp, H1_TS_D);
        strcat(fmtString, FMT_TS_D);
      }
      else
      {
        fprintf(out_fp, H1_TS);
        strcat(fmtString, FMT_TS);
      }

      fprintf(out_fp, H1_PID);
      strcat(fmtString, FMT_PID);
      if (verbose)
      {
        fprintf(out_fp, H1_P_HW_V);
        strcat(fmtString, FMT_P_HW_V);
      }
      fprintf(out_fp, H1_TAG);
      strcat(fmtString, FMT_TAG);

      /* Print second line of header */
      fprintf(out_fp, H2_SEQ);
      if (useTOD)
        fprintf(out_fp, H2_TS_T);
      else if (useLocal)
        fprintf(out_fp, H2_TS_D);
      else
        fprintf(out_fp, H2_TS);
      fprintf(out_fp, H2_PID);
      if (verbose)
        fprintf(out_fp, H2_P_HW_V);
      fprintf(out_fp, H2_TAG);
    }
#if defined(GPFS_LINUX) && defined(LXTRACE_USE_RELAY)
    //for multiple raw trace file, read first trace record
    if (nCPU > 1)
    {
      cpuHdr[curCPU].hdr.trMagic = LXTRACE_SEQ_MAGIC;
      for (i=curCPU+1; i<nCPU; i++)
      {
        if (ReadNextTrcHeader(raw_fp[i], &cpuHdr[i], i, prevSeq)<0)
          cpuHdr[i].hdr.trMagic = LXTRACE_NODATA_MAGIC;
      }
    }
#endif


    if (writeMode == LXTRACE_NB)
      prevSeq[0] = -1;

start_formatting:
    nWraps = 0; /* Reset the ReadNextTrcHeader wrap-found indicator. */
#ifdef SHARED_TRC_BUFF

    /* Process raw trace records. Start by reading the device header. */
    while (0 < ( h = PickNextTrcHeader(raw_fp, nCPU, cpuHdr, &newHdr,
                                                     hasWrapped, prevSeq)))
#else
    while (0 < ( h = PickNextTrcHeader(raw_fp, nCPU,
                                       cpuHdr, hasWrapped, prevSeq)))
#endif /* SHARED_TRC_BUFF */
    {
      #ifdef SHARED_TRC_BUFF
      if (writeMode == LXTRACE_NB)
      {
        /* Note we override the PickNextTrcHeader assignment here. */
        newHdr_t *h = &newHdr;
        recP = (trc_datahdr_t*) ( (char*)h + hdrSize[curCPU] - sizeof(trc_datahdr_t));
        DFP("rec.trHook 0x%X rec.trNArgs %d rec.trSPos %d rec.trSLen %d\n",
            recP->trHook, recP->trNArgs, recP->trSPos, recP->trSLen);

        if (h->h0.trLength < sizeof(trc_datahdr_t))
        {
          DP("lxtrace format: bad record length in %s at offset %d, trLength %d\n",
             rawTraceFileNameP, ftell(raw_fp[curCPU]), h->h0.trLength);
          rc = EINVAL;
          if (hasWrapped[curCPU])
          {
            DP("%s", "lxtrace format: restart from wrap-point\n");
            fprintf(out_fp, "*** BAD trace record header ***\n");
            goto do_wrap;
          }
          goto exit_format;
        }

        /* application_data[] holds the raw trace data that follows the headers. */
        int dataRead = 0;
        int applicationRead = h->h0.trLength - hdrSize[curCPU];
        if (0 > (dataRead = fread(application_data, 1,
                                  applicationRead, raw_fp[curCPU])))
        {
          DP("lxtrace format: could not read data from %s at offset %d, rc %d errno %d\n",
             rawTraceFileNameP, ftell(raw_fp[curCPU]), dataRead, errno);
          rc = errno;
          goto exit_format;
        }
        /* Since overwrite mode traces reserve space up to the nearest 64 bytes,
           we can have leftover data after a trace record. Skip over that. */
        int roundup =  (( h->h0.trLength + 63)/64) * 64 - h->h0.trLength;
        int rc = fseek(raw_fp[curCPU], roundup, SEEK_CUR);
        if(rc != 0)
          perror("lseek failed\n");
        DFP("Read %d bytes application data. Roundup %d\n", applicationRead, roundup);

        /* Get the formatting record for the trace record with this hookid */
        if (0 != GetFormatRecord(fmt_fp, (int)(recP->trHook & 0xFFFFFFFF),
                                 &compP, &fmtP, &i2nP, dataRead + hdrSize[curCPU]))
        {
          DP("lxtrace format: No formatting record found for hook %X\n",
             recP->trHook);
          fprintf(out_fp, "Unknown hook id 0x%08X\n", recP->trHook);
          DFP("raw data: %X %X %X %X %X\n",
              TRC_ARG(0), TRC_ARG(1), TRC_ARG(2), TRC_ARG(3), TRC_ARG(4));
          continue;
        }
        else
        {
          rc = formatOneTraceRecord(compP, fmtP, i2nP, application_data,
                                    &(newHdr.h0), NULL, recP, out_fp,
                                    fmtString, fmt_fp, verbose, sequence, useLocal);
          if(rc != 0)
          {
            DP("Error detected in formatting. Exit now\n");
            goto exit_format;
          }
          nRecordsProcessed += 1;
          continue;
        }
      }
#endif
        //trace records should not be missed if there's only one trace file
        //the per-CPU trace file may have this problem due to trace disbalance on each CPU
        if (nCPU ==1 && h->hdr.trMagic == LXTRACE_SEQ_MAGIC)
        {
          if (h->hdr.trSeq - lastSeq > 1 && lastSeq != 0)
          {
            fprintf(out_fp, "*** Missing %llu trace records ***\n", h->hdr.trSeq - lastSeq - 1 );
            lostRecords += ( h->hdr.trSeq - lastSeq - 1 );
          }
          lastSeq = h->hdr.trSeq;
        }

        recP = (trc_datahdr_t*) ( (char*)h + hdrSize[curCPU] - sizeof(trc_datahdr_t));
        DFP("rec.trHook 0x%X rec.trNArgs %d rec.trSPos %d rec.trSLen %d\n",
            recP->trHook, recP->trNArgs, recP->trSPos, recP->trSLen);

        if (h->hdr.trLength < sizeof(trc_datahdr_t))
        {
          DP("lxtrace format: bad record length in %s at offset %d, trLength %d\n",
             rawTraceFileNameP, ftell(raw_fp[curCPU]), h->hdr.trLength);
          rc = EINVAL;
          if (hasWrapped[curCPU])
          {
            DP("%s", "lxtrace format: restart from wrap-point\n");
            fprintf(out_fp, "*** BAD trace record header ***\n");
            goto do_wrap;
          }
          goto exit_format;
        }

        /* application_data[] holds the raw trace data that follows the headers. */
        int dataRead = 0;

        if (0 > (dataRead = fread(application_data, 1,
                                  h->hdr.trLength - sizeof(trc_datahdr_t), raw_fp[curCPU])))
        {
          DP("lxtrace format: could not read data from %s at offset %d, rc %d errno %d\n",
             rawTraceFileNameP, ftell(raw_fp[curCPU]), dataRead, errno);
          rc = errno;
          goto exit_format;
        }

        DFP("read %d bytes of application trace data\n", dataRead);

        /* If this is the first trace record, consider this time zero,
           unless the -T option was specified */
        if (tzero.tv_sec == 0  &&  !useTOD)
          tzero = h->hdr.trTime;

        /* Convert trace time to time since first trace record unless
           the -T option was specified */
        if (!useTOD)
        {
          h->hdr.trTime.tv_sec -= tzero.tv_sec;
          if (h->hdr.trTime.tv_usec >= tzero.tv_usec)
            h->hdr.trTime.tv_usec -= tzero.tv_usec;
          else
          {
            h->hdr.trTime.tv_sec--;
            h->hdr.trTime.tv_usec += 1000000 - tzero.tv_usec;
          }
        }
        /* Get the formatting record for the trace record with this hookid */
        if (0 != GetFormatRecord(fmt_fp, (int)(recP->trHook & 0xFFFFFFFF),
                                 &compP, &fmtP, &i2nP, dataRead + hdrSize[curCPU]))
        {
          DP("lxtrace format: No formatting record found for hook %X\n",
             recP->trHook);
          fprintf(out_fp, "Unknown hook id 0x%08X\n", recP->trHook);
          DFP("raw data: %X %X %X %X %X\n",
              TRC_ARG(0), TRC_ARG(1), TRC_ARG(2), TRC_ARG(3), TRC_ARG(4));

        }
        else
        {
          #ifdef SHARED_TRC_BUFF
          rc = formatOneTraceRecord(compP, fmtP, i2nP, application_data,
                                    NULL, h, recP, out_fp,
                                    fmtString, fmt_fp, verbose, sequence, useLocal);
          #else
          rc = formatOneTraceRecord(compP, fmtP, i2nP, application_data,
                                    h, recP, out_fp,
                                    fmtString, fmt_fp, verbose, sequence, useLocal);
          #endif
          if(rc != 0)
          {
            DP("Error detected in formatting. Exit now\n");
            goto exit_format;
          }
          nRecordsProcessed += 1;

        } /* end processing this trace record */
    } /* while there are more unformatted trace records to process */

    if (hasWrapped[curCPU])
    {
do_wrap:
      DFP("Rewinding buffer!\n");
      rewind(raw_fp[curCPU]); /* On a wrapped file, need to go back to the top. */
      hasWrapped[curCPU] = false;  /* Clear this so we don't do this more than once. */
      DFP("%s", "lxtrace format: done with wrap-point to EOF.  Going back to record 0\n");
      fileWrapped++;
      goto start_formatting;
    }

    rc = 0;

exit_format:

    /* Lost records count (primarily for per-cpu traces). The
       rest of the associated trace metadata is at the top of the file */
    if (out_fp != NULL)
      fprintf(out_fp, FMT_LOST_COUNT, lostRecords);

    /* print hash lookup statistics */
    DFP("Stats: hook lookup %6d check %7d (%.1f)\n",
        nHookLookup, nHookCheck, (double)nHookCheck/(double)nHookLookup);
    DFP("        i2n lookup %6d check %7d (%.1f) fail %d\n",
        nI2NLookup, nI2NCheck, (double)nI2NCheck/(double)nI2NLookup, nI2NFail);

    DFP("Wrap point: %d.%06d-%d.%06d (delta %d.%06d)\n",
        wrap_prev.tv_sec, wrap_prev.tv_usec, wrap_next.tv_sec, wrap_next.tv_usec,
        (wrap_prev.tv_sec-wrap_next.tv_sec), (wrap_prev.tv_usec-wrap_next.tv_usec));
#ifdef SHARED_TRC_BUFF
    /* Report how we did cycle counter conversions */
    if (tscSpeed == -1 && writeMode == LXTRACE_NB)
      fprintf(out_fp, dynamicTODMsg);
    else if (writeMode == LXTRACE_NB)
      fprintf(out_fp, staticTODMsg, tscSpeed);

    if (!showCycleCountOnly)
    {
      int prevCPUSpeed = -1;
      int changeFound = 0;

      for(i = 0; i < currTODCount; i++)
      {
        fprintf(out_fp, "Detected CPU running at %f MHz at %lld.%lld (%lld).\n", TODList[i]->cpuSpeed,
                TODList[i]->timestamp.tv_sec, TODList[i]->timestamp.tv_usec, TODList[i]->cycleCount);
        if(prevCPUSpeed != TODList[i]->cpuSpeed)
        {
          changeFound++;
          if (changeFound  >= 2)
          {
            fprintf(out_fp, "Note that this is a change in CPU speed. (if prior TOD times are with in a second of each other this is a spurious error)\n");
          }
        }
        prevCPUSpeed= TODList[i]->cpuSpeed;
      }
    }
#endif /* SHARED_TRC_BUFF */
    if (genStats)
    {
      int ctr;
      qsort(hookHashP, HOOK_HASH_TABLE_SIZE-1, sizeof(fmtDesc_t *), hookhashComparator);

      for (ctr = 0; ctr < HOOK_HASH_TABLE_SIZE-1; ctr++)
      {
        if (hookHashP[ctr] != NULL)
        {
          if (hookHashP[ctr]->numHits > 0)
          {
            fprintf(stat_fp, "hookid: 0x%X fmtP: %s     Hits: %d Total space: %lld bytes Average: %.2f Largest single entry: %d\n     Percent of uncompressed trcfile used: %.2f%% Percent of total traces: %.2f%%\n",
               hookHashP[ctr]->hookid,hookHashP[ctr]->fmtP,
               hookHashP[ctr]->numHits,hookHashP[ctr]->totalSpace,
               (double)hookHashP[ctr]->totalSpace/hookHashP[ctr]->numHits,hookHashP[ctr]->largestEntry,
               (double)hookHashP[ctr]->totalSpace/totalTrcfileProcessedSize*100,
               (double)hookHashP[ctr]->numHits/nRecordsProcessed*100);
          }
        }
      }
      fprintf(stat_fp, "Uncompressed trcfile space: %lld bytes. Compressed: %lld bytes. %lld traces seen\n",
              totalTrcfileProcessedSize, totalCompSize,nRecordsProcessed);
      int elapsed = stopTime - startTime;
      if(elapsed == 0)
      {
        fprintf(stat_fp, "Elapsed time is less than 1 second. Assume it's 1 second for calculations\n");
        elapsed = 1;
      }
      fprintf(stat_fp, "Traces cover %d seconds. Trace throughput was %.1f traces/second\n",
         elapsed, (double)nRecordsProcessed/elapsed);
      fprintf(stat_fp, "Uncompressed data: %.1f KB/sec, compressed data: %.1f KB/sec\n",
         (double)totalTrcfileProcessedSize/elapsed/1000, (double)totalCompSize/elapsed/1000);
      fprintf(stat_fp, "Nota Bene:\n");
      fprintf(stat_fp, "  Each trace entry comes with a header of %d bytes\n",hdrSize[0]);
      fprintf(stat_fp, "  Hits ==> total number of times this hookid was seen\n");
      fprintf(stat_fp, "  Total space ==> total space in the trcfile used by all traces with this hookid\n");
      fprintf(stat_fp, "  Average ==> total space / hits\n");
    }
    /* cleanup */
    if (formatFileP) free(formatFileP);
    if (fmtDescSpaceP) free(fmtDescSpaceP);
    for (i=0; i<nCPU; i++)
      if (raw_fp[i] != NULL) fclose(raw_fp[i]);
    if (fmt_fp != NULL) fclose(fmt_fp);
    if (out_fp != NULL)
    {
      fflush(out_fp);
      fclose(out_fp);
    }

  } /* "format" option */


  /* Display the state information associated with the device. */
  else if (strcmp(argv[1], "dump") == 0)
  {
    int fd;
#define DUMP_LEN 512
    char dumpBufP[DUMP_LEN];
    #ifdef SHARED_TRC_BUFF
    fd = open(TRC_DEVICE, O_RDWR);
#else
fd = open(TRC_DEVICE, O_RDWR);
#endif
    if (fd < 0)
    {
      DP("lxtrace dump: device open failed errno %d\n", errno);
      rc = ENODEV;
      goto exit;
    }

#if defined(GPFS_LINUX) && defined(LXTRACE_USE_RELAY)
    if ( RelayDump() < 0 )
#endif
    {

      /* Dump the trace device information */
      args.arg1 = DUMP_LEN;
#if defined(GPFS_LINUX)
      args.arg2 = (long)dumpBufP;
#elif defined(GPFS_SOLARIS)
      args.arg2 = (uint64_t)dumpBufP;
#endif
      if (0 != (rc = ioctl(fd, trc_dump, &args)))
      {
        DP("lxtrace dump: device dump returns rc %d errno %d \n", rc, errno);
      }
      else
      {
        printf("%s\n", dumpBufP);
      }
    }

exit_dump:
    if (fd >= 0)
      close(fd);
  }

  /* Flush any data that is currently buffered in the device. */
  else if (strcmp(argv[1], "fsync") == 0)
  {
    int fd;

    fd = open(TRC_DEVICE, O_RDONLY);
    if (fd < 0)
    {
      DP("lxtrace fsync: device open failed with errno %d\n", errno);
      rc = ENODEV;
      goto exit;
    }

#ifdef GPFS_SOLARIS
    if (0 != (rc = ioctl(fd, trc_fsynch, &args)))
#else
    if (0 != (rc = fsync(fd)))
#endif
    {
      DP("lxtrace fsync: device fsync returns rc %d errno %d\n", rc, errno);
    }

exit_fsync:
    if (fd >= 0)
      close(fd);
  }

  else if (strcmp(argv[1], "recycle") == 0 ||
           strcmp(argv[1], "off") == 0)
  {
    int fd;
    key_t key;
    int msgQueue = -1;
    int retry = 0;
    trc_msg msg;
    struct stat buf;
    time_t mtime;
#ifdef SHARED_TRC_BUFF
    Int64 bufferSize = 0;
    Int64 currentBufferOffset = 0;
    char *outFileP = NULL;
    for (i = 2; i < argc; i++)
    {
      if (strcmp(argv[i], "-w") == 0)
        writeMode = LXTRACE_NB;
      else if (strcmp(argv[i], "-s") == 0  &&  i+1<argc)
      {
        i += 1;
        /* Save the output filesize.  If the size is explicitly given as zero,
           do not wrap. */
        bufferSize = atoll(argv[i]);
      }
      else
        outFileP = argv[i];
    }
    if (writeMode == LXTRACE_NB)
    {
      char outFileName[256];
      if(bufferSize != 0)
      {
        DP("Don't provide a bufferSize for overwrite mode; we query the kernel\n");
        goto exit_recycle;
      }
      if(outFileP == NULL)
      {
        DP("You must provide the outfile\n");
        goto exit_recycle;
      }
      if (strcmp(argv[1], "recycle") == 0)
        sprintf(outFileName, "%s.recycle.cpu0", outFileP);
      else
        sprintf(outFileName, "%s.cpu0", outFileP);
      fd = open(TRC_DEVICE, O_RDWR);
      if (fd < 0)
      {
        DP("Could not open %s\n",TRC_DEVICE);
        rc = ENODEV;
        goto exit_recycle;
      }
      /* Get the buffer size from the kernel. If the trace buffer has
         never been filled, currentBufferOffset is smaller than bufferSize,
         so we only need that much data in the trcfile. */
      args.arg1 = (long)&bufferSize;
      args.arg2 = (long)&currentBufferOffset;
      rc = ioctl(fd, trc_size_query, &args);
      if(rc != 0)
      {
        DFP("Unable to get trace buffer size from kernel. rc %d\n", rc);
        goto exit_recycle;
      }
      if(bufferSize <= 0)
      {
        DFP("trc_size_query ioctl returned invalid result %d\n",bufferSize);
        rc = ENOMEM;
        goto exit_recycle;
      }
      /* Map the shared buffer to get the data. Tracing has been quiesced
         in the daemon/kernel so it's in a consistent state. */
      char *bufferP = (char *) mmap( NULL, bufferSize,
                                     PROT_WRITE | PROT_READ,
                                     MAP_SHARED,
                                     fd, 0);
      if(bufferP == (char *)MAP_FAILED)
      {
        DFP("Could not map memory, errno %d\n", errno);
        if(errno == EINVAL)
          DP("Mismatch in tracedevOverwriteBufferSize. You requested %lld. Please stop and restart traces\n", bufferSize);
        rc = ENOMEM;
        goto exit_recycle;
      }

      int hasWrapped = *(bufferP + sizeof(Int64));
      int openFlags = O_WRONLY|O_CREAT|O_DIRECT|O_TRUNC;
      int dumpHandle = open(outFileName, openFlags, 0666);
      if (dumpHandle == -1 && errno == EINVAL)
      {
        /* For tmpfs O_DIRECT doesn't work. Use O_SYNC instead. */
        errno = 0;
        openFlags = O_WRONLY|O_CREAT|O_SYNC|O_TRUNC;
        dumpHandle = open(outFileName, openFlags, 0666);
      }
      if (dumpHandle == -1)
      {
        DP("Cannot open dump file. rc %d errno %d (%s)\n",dumpHandle, errno, strerror(errno));
        return -1;
      }
      else
      {
        /* Write out the trace buffer */
        Int64 bytesToWrite = bufferSize;
        int writeWithDIO = 1;
        if (!hasWrapped)
          bytesToWrite = currentBufferOffset;
        if(writeWithDIO)
        {
          /* using mmap with DIO seems fraught with peril. Writing the buffer out directly
             fails without an aligned buffer, so I create an aligned buffer and copy it out
             piecemeal. I ran some quick experiments and larger buffers than 64K don't seem to
             make it go faster. */
          char *alignedBufP;
          int SECT_SIZE = 4096;
          int bytesPerLoop = 65536; /* This seems to give decent performance */
          rc = posix_memalign((void **)&alignedBufP, SECT_SIZE, bytesPerLoop);
          if(rc != 0)
          {
            DP("E_NOMEM with posix_memalign. Abort DIO and try regular IO\n");
            writeWithDIO = 0;
          }
          else
          {
            Int64 yetToWrite = bytesToWrite;
            while(yetToWrite > 0)
            {
              Int64 bytesThisLoop = MIN(bytesPerLoop, yetToWrite);
              memcpy(alignedBufP, bufferP + (bytesToWrite - yetToWrite), bytesThisLoop);
              Int64 nBytesWritten = write(dumpHandle, alignedBufP,
                                          bytesThisLoop);
              if (nBytesWritten != bytesThisLoop)
              {
                DP("Error writing trace dump using DIO. Attempted to write %lld but wrote %lld. errno %d (%s)\n",
                   bytesToWrite, nBytesWritten, errno, strerror(errno) );
                writeWithDIO = 0;
                break;
              }
              yetToWrite -= bytesPerLoop;
            }
            close(dumpHandle);
          }
        }
        /* This is a seperate if() so we can drop into this section without a goto if posix_memalign fails */
        if(!writeWithDIO)
        {
          close(dumpHandle);
          int dumpHandle = open(outFileName, O_WRONLY|O_CREAT, 0666);
          if (dumpHandle == -1)
          {
            DP("Cannot open dump file\n");
            return -1;
          }
          Int64 nBytesWritten = write(dumpHandle, bufferP,
                                    bytesToWrite);
          if (nBytesWritten != bytesToWrite)
          {
            DP("Error writing trace dump with regular IO. Attempted to write %lld but wrote %lld. errno %d (%s)\n",
               bytesToWrite, nBytesWritten, errno, strerror(errno) );
          }
          close(dumpHandle);
        }
        if (strcmp(argv[1], "off") == 0 )
        {
          /* trc_end terminates tracing. */
          args.arg1 = 0;
          rc = ioctl(fd, trc_end, &args);
          if (rc < 0)
          {
            DP("lxtrace off: device end failed with errno %d\n", errno);
            rc = EINVAL;
            goto exit_recycle;
          }
        }
        return 0;
      }
    }
#endif
    msgQueue = msgGet(IPC_CREAT | 0666);
    if (msgQueue == -1)
    {
      DP("lxtrace %s: message queue creation failed with errno %d\n", argv[1], errno);
      rc = errno;
      goto exit;
    }
    DFP("lxtrace %s: message queue %d created\n", argv[1], msgQueue);

    fd = open(TRC_DEVICE, O_RDONLY);
    if (fd < 0)
    {
      rc = ENODEV;
      goto exit_recycle;
    }

    if (strcmp(argv[1], "recycle") == 0 )
    {
      if (0 != (rc = ioctl(fd, trc_recycle, NULL)))
      {
#ifdef SHARED_TRC_BUFF
        if (errno == EBUSY)
        {
          rc = EBUSY;
          DP("lxtrace recycle: Currently in overwrite mode. Switch to blocking mode\n");
          goto exit_recycle;
        }
#endif /* SHARED_TRC_BUFF */
        DP("lxtrace recycle: device recycle returns rc %d errno %d\n", rc, errno);
        goto exit_recycle;
      }
    }
    else
    {
      /* trc_end terminates tracing. */
      args.arg1 = 0;
      rc = ioctl(fd, trc_end, &args);
      if (rc < 0)
      {
        DP("lxtrace off: device end failed with errno %d\n", errno);
        rc = EINVAL;
        goto exit_recycle;
      }
    }

    if (argc > 2 && stat(argv[2], &buf) == 0)
    {
      mtime = buf.st_mtime;
      DFP("lxtrace %s: trace file %s mtime %ld\n", argv[1], argv[2], buf.st_mtime);
    }
    while(1)
    {
      rc = msgrcv(msgQueue, &msg, 0, LXTRACE_MTYPE, IPC_NOWAIT);
      if (rc == -1)
      {
        rc = errno;
        //no message, sleep for 1/10th second and retry
        if (errno == ENOMSG)
        {
          if (retry++ >= 200)
          {
            if (argc>2 && stat(argv[2], &buf) == 0)
            {
              DFP("lxtrace %s: trace file %s mtime %ld\n", argv[1], argv[2], buf.st_mtime);
              //trace file is still being updated, wait for another round
              if (buf.st_mtime > mtime)
              {
                mtime = buf.st_mtime;
                retry = 0;
                continue;
              }
              else
                DFP("lxtrace %s: trace file %s not updated in last 20 seconds\n", argv[1], argv[2]);
            }
            DFP("lxtrace %s: msgrcv timeout error %d\n", argv[1], errno);
            break;
          }
          else
            usleep(100000);
          DFP("lxtrace %s: msgrcv return %d\n", argv[1], errno);
        }
        else
        {
          DP("lxtrace %s: msgrcv error %d\n", argv[1], errno);
          break;
        }
      }
      //get desired message
      else
      {
        DFP("lxtrace %s: msgrcv succeed\n", argv[1]);
        break;
      }
    }
exit_recycle:
    msgctl(msgQueue, IPC_RMID, NULL);
    close(fd);
  }
  else
    Usage();
exit:
  return rc;
}
#endif //GPFS_WINDOWS
